from typing import Dict, List, Set, Callable, Literal
from decimal import Decimal
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from nexustrader.core.log import SpdLog
from nexustrader.base import ExchangeManager
from nexustrader.core.entity import TaskManager
from nexustrader.core.cache import AsyncCache
from nexustrader.base import ExecutionManagementSystem, PrivateConnector
from nexustrader.core.nautilius_core import MessageBus, UUID4
from nexustrader.schema import (
    BookL1,
    Trade,
    Kline,
    Order,
    OrderSubmit,
    InstrumentId,
    BaseMarket,
    AccountBalance,
)
from nexustrader.constants import (
    DataType,
    OrderSide,
    OrderType,
    TimeInForce,
    PositionSide,
    AccountType,
    SubmitType,
    ExchangeType,
)


class Strategy:
    def __init__(self):
        self.log = SpdLog.get_logger(
            name=type(self).__name__, level="DEBUG", flush=True
        )

        self._subscriptions: Dict[DataType, Dict[str, str] | Set[str]] = {
            DataType.BOOKL1: set(),
            DataType.TRADE: set(),
            DataType.KLINE: {},
        }

        self._initialized = False
        self._scheduler = AsyncIOScheduler()

    def _init_core(
        self,
        exchanges: Dict[ExchangeType, ExchangeManager],
        private_connectors: Dict[AccountType, PrivateConnector],
        cache: AsyncCache,
        msgbus: MessageBus,
        task_manager: TaskManager,
        ems: Dict[ExchangeType, ExecutionManagementSystem],
    ):
        if self._initialized:
            return

        self.cache = cache

        self._ems = ems
        self._task_manager = task_manager
        self._msgbus = msgbus
        self._private_connectors = private_connectors
        self._exchanges = exchanges
        self._msgbus.subscribe(topic="trade", handler=self.on_trade)
        self._msgbus.subscribe(topic="bookl1", handler=self.on_bookl1)
        self._msgbus.subscribe(topic="kline", handler=self.on_kline)

        self._msgbus.register(endpoint="pending", handler=self.on_pending_order)
        self._msgbus.register(endpoint="accepted", handler=self.on_accepted_order)
        self._msgbus.register(
            endpoint="partially_filled", handler=self.on_partially_filled_order
        )
        self._msgbus.register(endpoint="filled", handler=self.on_filled_order)
        self._msgbus.register(endpoint="canceling", handler=self.on_canceling_order)
        self._msgbus.register(endpoint="canceled", handler=self.on_canceled_order)
        self._msgbus.register(endpoint="failed", handler=self.on_failed_order)
        self._msgbus.register(
            endpoint="cancel_failed", handler=self.on_cancel_failed_order
        )

        self._msgbus.register(endpoint="balance", handler=self.on_balance)

        self._initialized = True

    def schedule(
        self,
        func: Callable,
        trigger: Literal["interval", "cron"] = "interval",
        **kwargs,
    ):
        """
        cron: run at a specific time second, minute, hour, day, month, year
        interval: run at a specific interval  seconds, minutes, hours, days, weeks, months, years
        """

        self._scheduler.add_job(func, trigger=trigger, **kwargs)

    def market(self, symbol: str) -> BaseMarket:
        instrument_id = InstrumentId.from_str(symbol)
        exchange = self._exchanges[instrument_id.exchange]
        return exchange.market[instrument_id.symbol]

    def amount_to_precision(
        self,
        symbol: str,
        amount: float,
        mode: Literal["round", "ceil", "floor"] = "round",
    ) -> Decimal:
        instrument_id = InstrumentId.from_str(symbol)
        ems = self._ems[instrument_id.exchange]
        return ems._amount_to_precision(instrument_id.symbol, amount, mode)

    def price_to_precision(
        self,
        symbol: str,
        price: float,
        mode: Literal["round", "ceil", "floor"] = "round",
    ) -> Decimal:
        instrument_id = InstrumentId.from_str(symbol)
        ems = self._ems[instrument_id.exchange]
        return ems._price_to_precision(instrument_id.symbol, price, mode)

    def create_order(
        self,
        symbol: str,
        side: OrderSide,
        type: OrderType,
        amount: Decimal,
        price: Decimal | None = None,
        time_in_force: TimeInForce | None = TimeInForce.GTC,
        position_side: PositionSide | None = None,
        account_type: AccountType | None = None,
        **kwargs,
    ) -> str:
        order = OrderSubmit(
            symbol=symbol,
            instrument_id=InstrumentId.from_str(symbol),
            submit_type=SubmitType.CREATE,
            side=side,
            type=type,
            amount=amount,
            price=price,
            time_in_force=time_in_force,
            position_side=position_side,
            kwargs=kwargs,
        )
        self._ems[order.instrument_id.exchange]._submit_order(order, account_type)
        return order.uuid

    def cancel_order(
        self, symbol: str, uuid: str, account_type: AccountType | None = None, **kwargs
    ) -> str:
        order = OrderSubmit(
            symbol=symbol,
            instrument_id=InstrumentId.from_str(symbol),
            submit_type=SubmitType.CANCEL,
            uuid=uuid,
            kwargs=kwargs,
        )
        self._ems[order.instrument_id.exchange]._submit_order(order, account_type)
        return order.uuid

    def create_twap(
        self,
        symbol: str,
        side: OrderSide,
        amount: Decimal,
        duration: int,
        wait: int,
        position_side: PositionSide | None = None,
        account_type: AccountType | None = None,
        **kwargs,
    ) -> str:
        order = OrderSubmit(
            symbol=symbol,
            instrument_id=InstrumentId.from_str(symbol),
            uuid=f"ALGO-{UUID4().value}",
            submit_type=SubmitType.TWAP,
            side=side,
            amount=amount,
            duration=duration,
            wait=wait,
            position_side=position_side,
            kwargs=kwargs,
        )
        self._ems[order.instrument_id.exchange]._submit_order(order, account_type)
        return order.uuid

    def cancel_twap(self, symbol: str, uuid: str, account_type: AccountType | None = None) -> str:
        order = OrderSubmit(
            symbol=symbol,
            instrument_id=InstrumentId.from_str(symbol),
            submit_type=SubmitType.CANCEL_TWAP,
            uuid=uuid,
        )
        self._ems[order.instrument_id.exchange]._submit_order(order, account_type)
        return order.uuid

    def subscribe_bookl1(self, symbols: List[str]):
        """
        Subscribe to level 1 book data for the given symbols.

        Args:
            symbols (List[str]): The symbols to subscribe to.
        """
        for symbol in symbols:
            self._subscriptions[DataType.BOOKL1].add(symbol)

    def subscribe_trade(self, symbols: List[str]):
        """
        Subscribe to trade data for the given symbols.

        Args:
            symbols (List[str]): The symbols to subscribe to.
        """
        for symbol in symbols:
            self._subscriptions[DataType.TRADE].add(symbol)

    def subscribe_kline(self, symbols: List[str], interval: str):
        """
        Subscribe to kline data for the given symbols.

        Args:
            symbols (List[str]): The symbols to subscribe to.
            interval (str): The interval of the kline data
        """
        for symbol in symbols:
            self._subscriptions[DataType.KLINE][symbol] = interval

    def on_trade(self, trade: Trade):
        pass

    def on_bookl1(self, bookl1: BookL1):
        pass

    def on_kline(self, kline: Kline):
        pass

    def on_pending_order(self, order: Order):
        pass

    def on_accepted_order(self, order: Order):
        pass

    def on_partially_filled_order(self, order: Order):
        pass

    def on_filled_order(self, order: Order):
        pass

    def on_canceling_order(self, order: Order):
        pass

    def on_canceled_order(self, order: Order):
        pass

    def on_failed_order(self, order: Order):
        pass

    def on_cancel_failed_order(self, order: Order):
        pass

    def on_balance(self, balance: AccountBalance):
        pass
    
        



from typing import Any, Dict, List
from typing import Literal, Optional
from decimal import Decimal

from nexustrader.schema import Order


def in_orders(orders: List[Order], method: str, params: Dict[str, Any]) -> bool:
    for order in orders:
        match method:
            case "place_limit_order":
                if (
                    order.symbol == params["symbol"]
                    and order.side == params["side"]
                    and order.amount == params["amount"]
                    and order.price == params["price"]
                    and order.type == "limit"
                ):
                    return True
            case "place_market_order":
                if (
                    order.symbol == params["symbol"]
                    and order.side == params["side"]
                    and order.amount == params["amount"]
                    and order.type == "market"
                ):
                    return True
            case "cancel_order":
                if (
                    order.symbol == params["symbol"]
                    and order.id == params["id"]
                    and order.status == "canceled"
                ):
                    return True


def parse_ccxt_order(res: Dict[str, Any], exchange: str) -> Order:
    raw = res.get("info", {})
    id = res.get("id", None)
    client_order_id = res.get("clientOrderId", None)
    timestamp = res.get("timestamp", None)
    symbol = res.get("symbol", None)
    type = res.get("type", None)  # market or limit
    side = res.get("side", None)  # buy or sell
    price = res.get("price", None)  # maybe empty for market order
    average = res.get("average", None)  # float everage filling price
    amount = res.get("amount", None)
    filled = res.get("filled", None)
    remaining = res.get("remaining", None)
    status = raw.get("status", None).lower()
    cost = res.get("cost", None)
    reduce_only = raw.get("reduceOnly", None)
    position_side = raw.get("positionSide", "").lower() or None  # long or short
    time_in_force = res.get("timeInForce", None)

    return Order(
        raw=raw,
        success=True,
        exchange=exchange,
        id=id,
        client_order_id=client_order_id,
        timestamp=timestamp,
        symbol=symbol,
        type=type,
        side=side,
        price=price,
        average=average,
        amount=amount,
        filled=filled,
        remaining=remaining,
        status=status,
        cost=cost,
        reduce_only=reduce_only,
        position_side=position_side,
        time_in_force=time_in_force,
    )


def parse_websocket_stream(
    event_data: Dict[str, Any],
    market_id: Dict[str, Any],
    market_type: Optional[Literal["spot", "swap"]] = None,
):
    event = event_data.get("e", None)
    match event:
        case "kline":
            """
            {
                'e': 'kline', 
                'E': 1727525244267, 
                's': 'BTCUSDT', 
                'k': {
                    't': 1727525220000, 
                    'T': 1727525279999, 
                    's': 'BTCUSDT', 
                    'i': '1m', 
                    'f': 5422081499, 
                    'L': 5422081624, 
                    'o': '65689.80', 
                    'c': '65689.70', 
                    'h': '65689.80', 
                    'l': '65689.70', 
                    'v': '9.027', 
                    'n': 126, 
                    'x': False, 
                    'q': '592981.58290', 
                    'V': '6.610', 
                    'Q': '434209.57800', 
                    'B': '0'
                }
            }
            """
            id = f"{event_data['s']}_{market_type}" if market_type else event_data["s"]
            market = market_id[id]
            event_data["s"] = market["symbol"]
            return event_data


def parse_user_data_stream(event_data: Dict[str, Any], market_id: Dict[str, Any]):
    event = event_data.get("e", None)
    match event:
        case "ORDER_TRADE_UPDATE":
            """
            {
                "e": "ORDER_TRADE_UPDATE", // Event type
                "T": 1727352962757,  // Transaction time
                "E": 1727352962762, // Event time
                "fs": "UM", // Event business unit. 'UM' for USDS-M futures and 'CM' for COIN-M futures
                "o": {
                    "s": "NOTUSDT", // Symbol
                    "c": "c-11WLU7VP1727352880uzcu2rj4ss0i", // Client order ID
                    "S": "SELL", // Side
                    "o": "LIMIT", // Order type
                    "f": "GTC", // Time in force
                    "q": "5488", // Original quantity
                    "p": "0.0084830", // Original price
                    "ap": "0", // Average price
                    "sp": "0", // Ignore
                    "x": "NEW", // Execution type
                    "X": "NEW", // Order status
                    "i": 4968510801, // Order ID
                    "l": "0", // Order last filled quantity
                    "z": "0", // Order filled accumulated quantity
                    "L": "0", // Last filled price
                    "n": "0", // Commission, will not be returned if no commission
                    "N": "USDT", // Commission asset, will not be returned if no commission
                    "T": 1727352962757, // Order trade time
                    "t": 0, // Trade ID
                    "b": "0", // Bids Notional
                    "a": "46.6067521", // Ask Notional
                    "m": false, // Is this trade the maker side?
                    "R": false, // Is this reduce only
                    "ps": "BOTH", // Position side
                    "rp": "0", // Realized profit of the trade
                    "V": "EXPIRE_NONE", // STP mode
                    "pm": "PM_NONE", 
                    "gtd": 0
                }
            }
            """
            if event_data := event_data.get("o", None):
                if (market := market_id.get(event_data["s"], None)) is None:
                    id = f"{event_data['s']}_swap"
                    market = market_id[id]

                if (type := event_data["o"].lower()) == "market":
                    cost = float(event_data.get("l", "0")) * float(
                        event_data.get("ap", "0")
                    )
                elif type == "limit":
                    price = float(event_data.get("ap", "0")) or float(
                        event_data.get("p", "0")
                    )  # if average price is 0 or empty, use price
                    cost = float(event_data.get("l", "0")) * price

                return Order(
                    raw=event_data,
                    success=True,
                    exchange="binance",
                    id=event_data.get("i", None),
                    client_order_id=event_data.get("c", None),
                    timestamp=event_data.get("T", None),
                    symbol=market["symbol"],
                    type=type,
                    side=event_data.get("S", "").lower(),
                    status=event_data.get("X", "").lower(),
                    price=event_data.get("p", None),
                    average=event_data.get("ap", None),
                    last_filled_price=event_data.get("L", None),
                    amount=event_data.get("q", None),
                    filled=event_data.get("z", None),
                    last_filled=event_data.get("l", None),
                    remaining=Decimal(event_data["q"]) - Decimal(event_data["z"]),
                    fee=event_data.get("n", None),
                    fee_currency=event_data.get("N", None),
                    cost=cost,
                    last_trade_timestamp=event_data.get("T", None),
                    reduce_only=event_data.get("R", None),
                    position_side=event_data.get("ps", "").lower(),
                    time_in_force=event_data.get("f", None),
                )

        case "ACCOUNT_UPDATE":
            """
            {
                "e": "ACCOUNT_UPDATE", 
                "T": 1727352914268, 
                "E": 1727352914274, 
                "fs": "UM", 
                "a": {
                    "B": [
                        {"a": "USDT", "wb": "0.07147421", "cw": "0.07147421", "bc": "0"}, 
                        {"a": "BNB", "wb": "0.01993701", "cw": "0.01993701", "bc": "0"}
                    ], 
                    "P": [
                        {
                            "s": "BOMEUSDT", 
                            "pa": "-2760", 
                            "ep": "0.00724500", 
                            "cr": "0", 
                            "up": "-0.00077280", 
                            "ps": "BOTH", 
                            "bep": 0.0072436959
                        }
                    ], 
                "m": "ORDER"
                }
            }
            """
            positions = []
            for position in event_data["a"]["P"]:
                if (market := market_id.get(position["s"], None)) is None:
                    id = f"{position['s']}_swap"
                    market = market_id[id]
                position["s"] = market["symbol"]
                positions.append(position)
            event_data["a"]["P"] = positions
            return event_data

        case "balanceUpdate":
            """
            {
                "e": "balanceUpdate", 
                "E": 1727320813969, 
                "a": "BNB", 
                "d": "-0.01000000", 
                "U": 1495297874797, 
                "T": 1727320813969
            }
            """
            return event_data

        case "executionReport":
            """
            {
                "e": "executionReport", // Event type
                "E": 1727353057267, // Event time
                "s": "ORDIUSDT", // Symbol
                "c": "c-11WLU7VP2rj4ss0i", // Client order ID 
                "S": "BUY", // Side
                "o": "MARKET", // Order type
                "f": "GTC", // Time in force
                "q": "0.50000000", // Order quantity
                "p": "0.00000000", // Order price
                "P": "0.00000000", // Stop price
                "g": -1, // Order list id
                "x": "TRADE", // Execution type
                "X": "PARTIALLY_FILLED", // Order status
                "i": 2233880350, // Order ID
                "l": "0.17000000", // last executed quantity
                "z": "0.17000000", // Cumulative filled quantity
                "L": "36.88000000", // Last executed price
                "n": "0.00000216", // Commission amount
                "N": "BNB", // Commission asset
                "T": 1727353057266, // Transaction time
                "t": 105069149, // Trade ID
                "w": false, // Is the order on the book?
                "m": false, // Is this trade the maker side?
                "O": 1727353057266, // Order creation time
                "Z": "6.26960000", // Cumulative quote asset transacted quantity
                "Y": "6.26960000", // Last quote asset transacted quantity (i.e. lastPrice * lastQty)
                "V": "EXPIRE_MAKER", // Self trade prevention Mode
                "I": 1495839281094 // Ignore
            }
            
            # Example of an execution report event for a partially filled market buy order
            {
                "e": "executionReport", // Event type
                "E": 1727353057267, // Event time
                "s": "ORDIUSDT", // Symbol
                "c": "c-11WLU7VP2rj4ss0i", // Client order ID 
                "S": "BUY", // Side
                "o": "MARKET", // Order type
                "f": "GTC", // Time in force
                "q": "0.50000000", // Order quantity
                "p": "0.00000000", // Order price
                "P": "0.00000000", // Stop price
                "g": -1, // Order list id
                "x": "TRADE", // Execution type
                "X": "PARTIALLY_FILLED", // Order status
                "i": 2233880350, // Order ID
                "l": "0.17000000", // last executed quantity
                "z": "0.34000000", // Cumulative filled quantity
                "L": "36.88000000", // Last executed price
                "n": "0.00000216", // Commission amount
                "N": "BNB", // Commission asset
                "T": 1727353057266, // Transaction time
                "t": 105069150, // Trade ID
                "w": false, // Is the order on the book?
                "m": false, // Is this trade the maker side?
                "O": 1727353057266, // Order creation time
                "Z": "12.53920000", // Cumulative quote asset transacted quantity
                "Y": "6.26960000", // Last quote asset transacted quantity (i.e. lastPrice * lastQty)
                "V": "EXPIRE_MAKER", // Self trade prevention Mode
                "I": 1495839281094 // Ignore
            }
            
            """
            id = f"{event_data['s']}_spot"
            market = market_id[id]

            return Order(
                raw=event_data,
                success=True,
                exchange="binance",
                id=event_data.get("i", None),
                client_order_id=event_data.get("c", None),
                timestamp=event_data.get("T", None),
                symbol=market["symbol"],
                type=event_data.get("o", "").lower(),
                side=event_data.get("S", "").lower(),
                status=event_data.get("X", "").lower(),
                price=event_data.get("p", None),
                average=event_data.get("ap", None),
                last_filled_price=event_data.get("L", None),
                amount=event_data.get("q", None),
                filled=event_data.get("z", None),
                last_filled=event_data.get("l", None),
                remaining=Decimal(event_data.get("q", "0"))
                - Decimal(event_data.get("z", "0")),
                fee=event_data.get("n", None),
                fee_currency=event_data.get("N", None),
                cost=event_data.get("Y", None),
                last_trade_timestamp=event_data.get("T", None),
                time_in_force=event_data.get("f", None),
            )

        case "outboundAccountPosition":
            """
            {
                "e": "outboundAccountPosition", 
                "E": 1727353873873, 
                "u": 1727353873873, 
                "U": 1495859325408, 
                "B": [
                    {
                        "a": "BNB", 
                        "f": 
                        "0.09971173", 
                        "l": 
                        "0.00000000"
                    }, 
                    {
                        "a": "USDT", 
                        "f": "6426.31521496", 
                        "l": "0.00000000"
                    }, 
                    {"a": 
                    "AVAX", 
                    "f": "3.00000000", 
                    "l": "0.00000000"
                    }
                ]
            }
            """
            return event_data



import asyncio
from decimal import Decimal
from typing import Dict
from nexustrader.constants import AccountType
from nexustrader.schema import OrderSubmit
from nexustrader.core.cache import AsyncCache
from nexustrader.core.nautilius_core import MessageBus
from nexustrader.core.entity import TaskManager
from nexustrader.core.registry import OrderRegistry
from nexustrader.exchange.okx import OkxAccountType
from nexustrader.exchange.okx.schema import OkxMarket
from nexustrader.base import ExecutionManagementSystem


class OkxExecutionManagementSystem(ExecutionManagementSystem):
    _market: Dict[str, OkxMarket]

    OKX_ACCOUNT_TYPE_PRIORITY = [
        OkxAccountType.DEMO,
        OkxAccountType.AWS,
        OkxAccountType.LIVE,
    ]
    
    def __init__(
        self,
        market: Dict[str, OkxMarket],
        cache: AsyncCache,
        msgbus: MessageBus,
        task_manager: TaskManager,
        registry: OrderRegistry,
    ):
        super().__init__(
            market=market,
            cache=cache,
            msgbus=msgbus,
            task_manager=task_manager,
            registry=registry,
        )
        self._okx_account_type: OkxAccountType = None

    def _build_order_submit_queues(self):
        for account_type in self._private_connectors.keys():
            if isinstance(account_type, OkxAccountType):
                self._order_submit_queues[account_type] = asyncio.Queue()
                break

    def _set_account_type(self):
        account_types = self._private_connectors.keys()
        for account_type in self.OKX_ACCOUNT_TYPE_PRIORITY:
            if account_type in account_types:
                self._okx_account_type = account_type
                break

    def _submit_order(
        self, order: OrderSubmit, account_type: AccountType | None = None
    ):
        if not account_type:
            account_type = self._okx_account_type
        self._order_submit_queues[account_type].put_nowait(order)

    def _get_min_order_amount(self, symbol: str, market: OkxMarket) -> Decimal:
        pass



from pathlib import Path
from typing import Literal
import sys
import traceback
import asyncio
import spdlog as spd
     

class SpdLog:
    """
    Log registration class responsible for creating and managing loggers.

    Features:
    - Supports multiple log levels
    - Structured log output (e.g., JSON format)
    - Captures and logs synchronous and asynchronous exceptions
    - Supports log rotation
    - Allows managing log settings via configuration files or environment variables
    """

    log_dir = Path(".log")
    log_dir_created = False
    loggers = {}
    async_mode = True
    error_logger = None
    sinks = None
    production_mode = False
    


    @classmethod
    def setup_error_handling(cls):
        cls.error_logger = cls.get_logger("Error", level="ERROR", flush=True)

        def handle_exception(exc_type, exc_value, exc_traceback):
            if issubclass(exc_type, KeyboardInterrupt):
                sys.__excepthook__(exc_type, exc_value, exc_traceback)
                return
            tb_str = "".join(
                traceback.format_exception(exc_type, exc_value, exc_traceback)
            )
            cls.error_logger.error(tb_str)

        sys.excepthook = handle_exception

        def handle_async_exception(loop, async_context):
            msg = async_context.get(
                "exception", async_context.get("message", "Unknown async exception")
            )
            if "exception" in async_context:
                exception = async_context["exception"]
                tb_str = "".join(
                    traceback.format_exception(
                        type(exception), exception, exception.__traceback__
                    )
                )
                cls.error_logger.error(tb_str)
            else:
                cls.error_logger.error(f"Caught async exception: {msg}")

        asyncio.get_event_loop().set_exception_handler(handle_async_exception)

    @classmethod
    def get_logger(
        cls,
        name: str,
        level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = "INFO",
        flush: bool = False,
    ) -> spd.Logger:
        """
        Get the logger with the specified name. If it doesn't exist, create a new logger.

        :param name: Logger name
        :param level: Log level
        :param flush: Whether to flush after each log entry
        :return: spdlog.Logger instance
        """
        if name not in cls.loggers:
            if not cls.log_dir_created:
                cls.log_dir.mkdir(parents=True, exist_ok=True)
                cls.log_dir_created = True
            if cls.production_mode:
                logger_instance = spd.SinkLogger(name=name, sinks=cls.sinks)
            else:
                logger_instance = spd.DailyLogger(
                    name=name,
                    filename=str(cls.log_dir / f"{name}.log"),
                    hour=0,
                    minute=0,
                    async_mode=cls.async_mode,
                )
            logger_instance.set_level(cls.parse_level(level))
            if flush:
                logger_instance.flush_on(cls.parse_level(level))
            cls.loggers[name] = logger_instance
        return cls.loggers[name]

    @classmethod
    def parse_level(
        cls, level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
    ) -> spd.LogLevel:
        """
        Parse the log level string to spdlog.LogLevel.

        :param level: Log level string
        :return: spdlog.LogLevel
        """
        levels = {
            "DEBUG": spd.LogLevel.DEBUG,
            "INFO": spd.LogLevel.INFO,
            "WARNING": spd.LogLevel.WARN,
            "ERROR": spd.LogLevel.ERR,
            "CRITICAL": spd.LogLevel.CRITICAL,
        }
        return levels[level]

    @classmethod
    def close_all_loggers(cls):
        """
        Close all loggers and release resources.
        """
        for logger in cls.loggers.values():
            logger.flush()
            logger.drop()

    @classmethod
    def initialize(cls, log_dir: str = ".log", async_mode: bool = True, setup_error_handlers: bool = True):
        """
        Initialize the log registry.

        :param setup_error_handlers: Whether to set up global exception handlers
        :param log_dir: Log directory
        :param async_mode: Whether to enable asynchronous mode
        """
        cls.log_dir = Path(log_dir)
        cls.async_mode = async_mode
        if setup_error_handlers:
            cls.setup_error_handling()
        if cls.production_mode:
            daily_sink = spd.daily_file_sink_mt(filename=str(log_dir / "log.log"), rotation_hour=0, rotation_minute=0)
            daily_sink.set_level(spd.LogLevel.INFO)
            
            stdout_sink = spd.stdout_color_sink_mt()    
            stdout_sink.set_level(spd.LogLevel.DEBUG)
            
            cls.sinks = [
                daily_sink,
                stdout_sink,
            ]
            
    @classmethod
    def __del__(cls):
        cls.close_all_loggers()


if __name__ == "__main__":
    SpdLog.initialize(setup_error_handlers=False)  # You should initialize the logger in the main thread
    logger = SpdLog.get_logger("test", level="DEBUG", flush=True)
    logger.debug("This is a debug message")
    logger.info("This is an info message")
    logger.warn("This is a warning message")
    logger.error("This is an error message")
    logger.critical("This is a critical message")
    raise ValueError("This is a test exception")



from decimal import Decimal
from typing import Dict, List, Tuple, Any
from typing import Optional
from msgspec import Struct, field
from nexustrader.core.nautilius_core import UUID4
from nexustrader.constants import (
    OrderSide,
    OrderType,
    TimeInForce,
    OrderStatus,
    PositionSide,
    InstrumentType,
    ExchangeType,
    SubmitType,
    AlgoOrderStatus,
)


class InstrumentId(Struct):
    symbol: str
    exchange: ExchangeType
    type: InstrumentType

    @property
    def is_spot(self) -> bool:
        return self.type == InstrumentType.SPOT

    @property
    def is_linear(self) -> bool:
        return self.type == InstrumentType.LINEAR

    @property
    def is_inverse(self) -> bool:
        return self.type == InstrumentType.INVERSE

    @classmethod
    def from_str(cls, symbol: str):
        """
        BTCETH.BINANCE -> SPOT
        BTCUSDT-PERP.BINANCE -> LINEAR
        BTCUSD.BINANCE -> INVERSE
        BTCUSD-241227.BINANCE
        """
        symbol_prefix, exchange = symbol.split(".")

        # if numirical number in id, then it is a future
        if "-" in symbol_prefix:
            prefix, _ = symbol_prefix.split("-")
            if prefix.endswith("USD"):
                type = InstrumentType.INVERSE
            else:
                type = InstrumentType.LINEAR
        else:
            type = InstrumentType.SPOT

        return cls(symbol=symbol, exchange=ExchangeType(exchange.lower()), type=type)


class BookL1(Struct, gc=False):
    exchange: ExchangeType
    symbol: str
    bid: float
    ask: float
    bid_size: float
    ask_size: float
    timestamp: int


class BookL2(Struct):
    exchange: ExchangeType
    symbol: str
    bids: List[Tuple[float, float]]
    asks: List[Tuple[float, float]]
    timestamp: int


class Trade(Struct, gc=False):
    exchange: ExchangeType
    symbol: str
    price: float
    size: float
    timestamp: int


class Kline(Struct, gc=False):
    exchange: ExchangeType
    symbol: str
    interval: str
    open: float
    high: float
    low: float
    close: float
    volume: float
    timestamp: int


class MarkPrice(Struct, gc=False):
    exchange: ExchangeType
    symbol: str
    price: float
    timestamp: int


class FundingRate(Struct, gc=False):
    exchange: ExchangeType
    symbol: str
    rate: float
    timestamp: int
    next_funding_time: int


class IndexPrice(Struct, gc=False):
    exchange: ExchangeType
    symbol: str
    price: float
    timestamp: int


class OrderSubmit(Struct):
    symbol: str
    instrument_id: InstrumentId
    submit_type: SubmitType
    uuid: str = field(default_factory=lambda: UUID4().value)
    order_id: str | int | None = None
    side: OrderSide | None = None
    type: OrderType | None = None
    amount: Decimal | None = None
    price: Decimal | None = None
    time_in_force: TimeInForce | None = TimeInForce.GTC
    position_side: PositionSide | None = None
    duration: int | None = None
    wait: int | None = None
    kwargs: Dict[str, Any] = {}
    status: OrderStatus = OrderStatus.INITIALIZED


class Order(Struct):
    exchange: ExchangeType
    symbol: str
    status: OrderStatus
    id: Optional[str] = None
    uuid: Optional[str] = None
    amount: Optional[Decimal] = None
    filled: Optional[Decimal] = None
    client_order_id: Optional[str] = None
    timestamp: Optional[int] = None
    type: Optional[OrderType] = None
    side: Optional[OrderSide] = None
    time_in_force: Optional[TimeInForce] = None
    price: Optional[float] = None
    average: Optional[float] = None
    last_filled_price: Optional[float] = None
    last_filled: Optional[Decimal] = None
    remaining: Optional[Decimal] = None
    fee: Optional[Decimal] = None
    fee_currency: Optional[str] = None
    cost: Optional[Decimal] = None
    cum_cost: Optional[Decimal] = None
    reduce_only: Optional[bool] = None
    position_side: Optional[PositionSide] = None

    @property
    def success(self) -> bool:
        return self.status not in [OrderStatus.FAILED, OrderStatus.CANCEL_FAILED]

    @property
    def is_filled(self) -> bool:
        return self.status == OrderStatus.FILLED

    @property
    def is_canceled(self) -> bool:
        return self.status == OrderStatus.CANCELED

    @property
    def is_closed(self) -> bool:
        return self.status in [
            OrderStatus.FILLED,
            OrderStatus.CANCELED,
            OrderStatus.EXPIRED,
        ]

    @property
    def is_opened(self) -> bool:
        return self.status in [
            OrderStatus.PENDING,
            OrderStatus.CANCELING,
            OrderStatus.PARTIALLY_FILLED,
            OrderStatus.ACCEPTED,
        ]
    
    @property
    def on_flight(self) -> bool:
        return self.status in [
            OrderStatus.PENDING,
            OrderStatus.CANCELING,
        ]


class AlgoOrder(Struct):
    symbol: str
    uuid: str # start with "ALGO-"
    side: OrderSide
    amount: Decimal
    duration: int
    wait: int
    status: AlgoOrderStatus
    exchange: ExchangeType
    timestamp: int 
    orders: List[str] = field(default_factory=list) # [uuid1, uuid2, ...]
    position_side: PositionSide | None = None
    
    @property
    def is_running(self) -> bool:
        return self.status == AlgoOrderStatus.RUNNING
    
    @property
    def is_finished(self) -> bool:
        return self.status == AlgoOrderStatus.FINISHED
    
    @property
    def is_canceled(self) -> bool:
        return self.status == AlgoOrderStatus.CANCELED
    
    @property
    def is_failed(self) -> bool:
        return self.status == AlgoOrderStatus.FAILED
    
    @property
    def is_closed(self) -> bool:
        return self.status in [AlgoOrderStatus.CANCELED, AlgoOrderStatus.FAILED, AlgoOrderStatus.FINISHED]
    
    @property
    def is_opened(self) -> bool:
        return self.status in [AlgoOrderStatus.RUNNING, AlgoOrderStatus.CANCELING]

class Balance(Struct):
    """
    Buy BTC/USDT: amount = 0.01, cost: 600

    OrderStatus.INITIALIZED: BTC(free: 0.0, locked: 0.0) USDT(free: 1000, locked: 0)
    OrderStatus.PENDING: BTC(free: 0.0, locked: 0) USDT(free: 400, locked: 600) USDT.update_locked(600) USDT.update_free(-600)

    OrderStatus.PARTIALLY_FILLED: BTC(free: 0.005, locked: 0) USDT(free: 400, locked: 300) BTC.update_free(0.005) USDT.update_locked(-300)
    OrderStatus.FILLED: BTC(free: 0.01, locked: 0.0) USDT(free: 400, locked: 0) BTC.update_free(0.005) USDT.update_locked(-300)

    Buy BTC/USDT: amount = 0.01, cost: 200

    OrderStatus.INITIALIZED: BTC(free: 0.01, locked: 0.0) USDT(free: 400, locked: 0)
    OrderStatus.PENDING: BTC(free: 0.01, locked: 0.0) USDT(free: 200, locked: 200) USDT.update_locked(200) USDT.update_free(-200)
    OrderStatus.FILLED: BTC(free: 0.02, locked: 0.0) USDT(free: 200, locked: 0) BTC.update_free(0.01) USDT.update_locked(-200)

    Sell BTC/USDT: amount = 0.01, cost: 300
    OrderStatus.INITIALIZED: BTC(free: 0.02, locked: 0.0) USDT(free: 200, locked: 0)
    OrderStatus.PENDING: BTC(free: 0.01, locked: 0.01) USDT(free: 200, locked: 0) BTC.update_locked(0.01) BTC.update_free(-0.01)
    OrderStatus.PARTIALLY_FILLED: BTC(free: 0.01, locked: 0.005) USDT(free: 350, locked: 0) BTC.update_locked(-0.005) USDT.update_free(150)
    OrderStatus.FILLED: BTC(free: 0.01, locked: 0.0) USDT(free: 500, locked: 0) BTC.update_locked(-0.005) USDT.update_free(150)
    """

    asset: str
    free: Decimal = field(default=Decimal("0.0"))
    locked: Decimal = field(default=Decimal("0.0"))

    @property
    def total(self) -> Decimal:
        return self.free + self.locked


class AccountBalance(Struct):
    balances: Dict[str, Balance] = field(default_factory=dict)

    def _apply(self, balances: List[Balance]):
        for balance in balances:
            self.balances[balance.asset] = balance

    @property
    def balance_total(self) -> Dict[str, Decimal]:
        return {asset: balance.total for asset, balance in self.balances.items()}

    @property
    def balance_free(self) -> Dict[str, Decimal]:
        return {asset: balance.free for asset, balance in self.balances.items()}

    @property
    def balance_locked(self) -> Dict[str, Decimal]:
        return {asset: balance.locked for asset, balance in self.balances.items()}


class Precision(Struct):
    """
     "precision": {
      "amount": 0.0001,
      "price": 1e-05,
      "cost": null,
      "base": 1e-08,
      "quote": 1e-08
    },
    """

    amount: float | None = None
    price: float | None = None
    cost: float | None = None
    base: float | None = None
    quote: float | None = None


class LimitMinMax(Struct):
    """
    "limits": {
      "amount": {
        "min": 0.0001,
        "max": 1000.0
      },
      "price": {
        "min": 1e-05,
        "max": 1000000.0
      },
      "cost": {
        "min": 0.01,
        "max": 1000000.0
      }
    },
    """

    min: float | None
    max: float | None


class Limit(Struct):
    leverage: LimitMinMax = None
    amount: LimitMinMax = None
    price: LimitMinMax = None
    cost: LimitMinMax = None
    market: LimitMinMax = None


class MarginMode(Struct):
    isolated: bool | None
    cross: bool | None


class BaseMarket(Struct):
    """Base market structure for all exchanges."""

    id: str
    lowercaseId: str | None
    symbol: str
    base: str
    quote: str
    settle: str | None
    baseId: str
    quoteId: str
    settleId: str | None
    type: InstrumentType
    spot: bool
    margin: bool | None
    swap: bool
    future: bool
    option: bool
    index: bool | str | None
    active: bool
    contract: bool
    linear: bool | None
    inverse: bool | None
    subType: InstrumentType | None
    taker: float
    maker: float
    contractSize: float | None
    expiry: int | None
    expiryDatetime: str | None
    strike: float | str | None
    optionType: str | None
    precision: Precision
    limits: Limit
    marginModes: MarginMode
    created: int | None
    tierBased: bool | None
    percentage: bool | None
    # feeSide: str  # not supported by okx exchanges


"""
class Position(Struct):

    one-way mode:
    > order (side: buy) -> side: buy | pos_side: net/both | reduce_only: False [open long position]
    > order (side: sell) -> side: sell | pos_side: net/both | reduce_only: False [open short position]
    > order (side: buy, reduce_only=True) -> side: buy | pos_side: net/both | reduce_only: True [close short position]
    > order (side: sell, reduce_only=True) -> side: sell | pos_side: net/both | reduce_only: True [close long position]

    hedge mode:
    > order (side: buy, pos_side: long) -> side: buy | pos_side: long | reduce_only: False [open long position]
    > order (side: sell, pos_side: short) -> side: sell | pos_side: short | reduce_only: False [open short position]
    > order (side: sell, pos_side: long) -> side: sell | pos_side: long | reduce_only: True [close long position]
    > order (side: buy, pos_side: short) -> side: buy | pos_side: short | reduce_only: True [close short position]

    
"""


class Position(Struct):
    symbol: str
    exchange: ExchangeType
    signed_amount: Decimal = Decimal("0")
    entry_price: float = 0
    side: Optional[PositionSide] = None
    unrealized_pnl: float = 0
    realized_pnl: float = 0

    @property
    def amount(self, contract_size: Decimal = Decimal("1")) -> Decimal:
        return abs(self.signed_amount) * contract_size

    @property
    def is_open(self) -> bool:
        return self.amount != 0

    @property
    def is_closed(self) -> bool:
        return not self.is_open

    @property
    def is_long(self) -> bool:
        return self.side == PositionSide.LONG

    @property
    def is_short(self) -> bool:
        return self.side == PositionSide.SHORT


# class SpotPosition(Position):
#     @property
#     def amount(self) -> Decimal:
#         return abs(self.signed_amount)

#     _last_order_filled: Dict[str, Decimal] = field(default_factory=dict)

#     def _calculate_fill_delta(self, order: Order) -> Decimal:
#         """
#         calculate the fill delta of the order, since filled in order is cumulative,
#         we need to calculate the delta of the order
#         """
#         previous_fill = self._last_order_filled.get(order.uuid, Decimal("0"))
#         current_fill = order.filled
#         fill_delta = current_fill - previous_fill
#         if order.status in (OrderStatus.FILLED, OrderStatus.CANCELED):
#             self._last_order_filled.pop(order.uuid, None)
#         else:
#             self._last_order_filled[order.uuid] = order.filled
#         return fill_delta

#     def _apply(self, order: Order):
#         if not order.last_filled:
#             fill_delta = self._calculate_fill_delta(order)
#         else:
#             fill_delta = order.last_filled

#         if order.side == OrderSide.BUY:
#             self.signed_amount += fill_delta
#         elif order.side == OrderSide.SELL:
#             self.signed_amount -= fill_delta


# class Position(Struct):
#     symbol: str
#     exchange: ExchangeType
#     side: Optional[PositionSide] = None
#     signed_amount: Decimal = Decimal("0")
#     entry_price: float = 0
#     unrealized_pnl: float = 0
#     realized_pnl: float = 0
#     _last_order_filled: Dict[str, Decimal] = field(default_factory=dict)

#     @property
#     def amount(self) -> Decimal:
#         return abs(self.signed_amount)

#     @property
#     def is_open(self) -> bool:
#         return self.amount != 0

#     @property
#     def is_closed(self) -> bool:
#         return not self.is_open

#     @property
#     def is_long(self) -> bool:
#         return self.side == PositionSide.LONG

#     @property
#     def is_short(self) -> bool:
#         return self.side == PositionSide.SHORT

#     def _calculate_fill_delta(self, order: Order) -> Decimal:
#         """
#         calculate the fill delta of the order, since filled in order is cumulative,
#         we need to calculate the delta of the order
#         """
#         previous_fill = self._last_order_filled.get(order.uuid, Decimal("0"))
#         current_fill = order.filled
#         fill_delta = current_fill - previous_fill
#         if order.status in (OrderStatus.FILLED, OrderStatus.CANCELED):
#             self._last_order_filled.pop(order.uuid, None)
#         else:
#             self._last_order_filled[order.uuid] = order.filled
#         return fill_delta

#     def _calculate_pnl(self, current_price: float, amount: Decimal) -> float:
#         """Calculate PNL based on position side and current price"""
#         if self.is_long:
#             return float(amount) * (current_price - self.entry_price)
#         elif self.is_short:
#             return float(amount) * (self.entry_price - current_price)
#         return 0.0

#     def apply(self, order: Order):
#         if order.position_side == PositionSide.FLAT:
#             if not order.last_filled:
#                 fill_delta = self._calculate_fill_delta(order)
#             else:
#                 fill_delta = order.last_filled

#             if (self.signed_amount > 0 and order.side == OrderSide.SELL) or (
#                 self.signed_amount < 0 and order.side == OrderSide.BUY
#             ):
#                 close_amount = min(
#                     abs(self.signed_amount), fill_delta
#                 )  # 平仓数量最大不超过当前持仓数量
#                 remaining_amount = fill_delta - close_amount  # 剩余数量
#                 self._close_position(order, close_amount)
#                 if remaining_amount > 0:
#                     self._open_position(order, remaining_amount)
#             else:
#                 self._open_position(order, fill_delta)
#         else:
#             pass

#     def _update_position(
#         self, current_price: float, amount: Decimal, buy_sell: OrderSide
#     ):
#         amount_change = amount if buy_sell == OrderSide.BUY else -amount

#         # 1. update entry price
#         self.entry_price = (
#             (
#                 self.entry_price * float(self.signed_amount)
#                 + current_price * float(amount_change)
#             )
#             / float(self.signed_amount + amount_change)
#             if self.signed_amount + amount_change != 0
#             else 0
#         )

#         # 2. update signed amount
#         self.signed_amount += amount_change

#         # 3. update unrealized pnl
#         self.unrealized_pnl = self._calculate_pnl(current_price, self.amount)

#     def _close_position(self, order: Order, close_amount: Decimal):
#         if order.side == OrderSide.BUY:
#             if not self.is_short:
#                 warnings.warn(f"Cannot close short position with {self.side}")
#         elif order.side == OrderSide.SELL:
#             if not self.is_long:
#                 warnings.warn(f"Cannot close long position with {self.side}")

#         price = order.average or order.price
#         self._update_position(price, close_amount, order.side)
#         self.realized_pnl += self._calculate_pnl(price, close_amount)

#         if self.signed_amount == 0:
#             self.side = None
#             self.entry_price = 0
#             self.unrealized_pnl = 0

#     def _open_position(self, order: Order, open_amount: Decimal):
#         if order.side == OrderSide.BUY:
#             if not self.side:
#                 self.side = PositionSide.LONG
#             else:
#                 if not self.is_long:
#                     warnings.warn(f"Cannot open long position with {self.side}")

#         elif order.side == OrderSide.SELL:
#             if not self.side:
#                 self.side = PositionSide.SHORT
#             else:
#                 if not self.is_short:
#                     warnings.warn(f"Cannot open short position with {self.side}")

#         price = order.average or order.price
#         self._update_position(price, open_amount, order.side)



import hashlib
import hmac
import orjson

from typing import Any, Dict
from urllib.parse import urljoin, urlencode


from nexustrader.core.log import SpdLog
from nexustrader.constants import OrderSide, OrderType
from nexustrader.exchange.binance.error import BinanceClientError, BinanceServerError
from nexustrader.core.nautilius_core import LiveClock, HttpClient, HttpMethod, HttpResponse


class BinanceHttpClient:
    def __init__(
        self,
        api_key: str,
        secret: str,
        testnet: bool = False,
    ):
        self._api_key = api_key
        self._secret = secret
        self._testnet = testnet
        self._clock = LiveClock()
        self._log = SpdLog.get_logger(type(self).__name__, level="INFO", flush=True)

        self._headers = {
            "Content-Type": "application/json",
            "User-Agent": "TradingBot/1.0",
            "X-MBX-APIKEY": api_key,
        }

        self._client = HttpClient()

    def _generate_signature(self, query: str) -> str:
        signature = hmac.new(
            self._secret.encode("utf-8"), query.encode("utf-8"), hashlib.sha256
        ).hexdigest()
        return signature

    async def _fetch(
        self,
        method: HttpMethod,
        base_url: str,
        endpoint: str,
        payload: Dict[str, Any] = None,
        signed: bool = False,
    ):
        url = urljoin(base_url, endpoint)
        payload = payload or {}
        payload["timestamp"] = self._clock.timestamp_ms()
        payload = urlencode(payload)

        if signed:
            signature = self._generate_signature(payload)
            payload += f"&signature={signature}"

        url += f"?{payload}"
        self._log.debug(f"Request: {url}")

        response: HttpResponse = await self._client.request(
            method=method,
            url=url,
            headers=self._headers,
        )

        if 400 <= response.status < 500:
            raise BinanceClientError(
                status=response.status,
                message=orjson.loads(response.body) if response.body else None,
                headers=response.headers,
            )
        elif response.status >= 500:
            raise BinanceServerError(
                status=response.status,
                message=orjson.loads(response.body) if response.body else None,
                headers=response.headers,
            )
        return orjson.loads(response.body)

    async def post_fapi_v1_order(
        self,
        symbol: str,
        side: OrderSide,
        type: OrderType,
        **kwargs,
    ):
        if self._testnet:
            base_url = "https://testnet.binancefuture.com"
        else:
            base_url = "https://fapi.binance.com"

        end_point = "/fapi/v1/order"

        payload = {
            "symbol": symbol,
            "side": side.value,
            "type": type.value,
            **kwargs,
        }

        return await self._fetch(
            HttpMethod.POST, base_url, end_point, payload, signed=True
        )

    async def post_papi_v1_um_order(
        self,
        symbol: str,
        side: OrderSide,
        type: OrderType,
        **kwargs,
    ):
        base_url = "https://papi.binance.com"
        end_point = "/papi/v1/um/order"

        payload = {
            "symbol": symbol,
            "side": side.value,
            "type": type.value,
            **kwargs,
        }

        return await self._fetch(
            HttpMethod.POST, base_url, end_point, payload, signed=True
        )



import asyncio
from typing import Callable, Any, Optional
from nautilus_trader.common.component import LiveClock
from nautilus_trader.adapters.okx.websocket.client import OKXWebsocketClient
from nautilus_trader.adapters.okx.common.enums import OKXWsBaseUrlType, OKXBarSize
from nexustrader.exchange.okx import OkxAccountType


class OkxWSClient:
    """OKX WebSocket client manager with business logic."""

    def __init__(
        self,
        handler: Callable[..., Any],
        api_key: Optional[str] = None,
        api_secret: Optional[str] = None,
        passphrase: Optional[str] = None,
        is_demo: bool = False,
        account_type: Optional[OkxAccountType] = None,  # compatible with old connector
    ):
        self._clock = LiveClock()
        self._loop = asyncio.get_event_loop()
        self._handler = handler
        self._api_key = api_key
        self._api_secret = api_secret
        self._passphrase = passphrase
        self._is_demo = is_demo
        self._reconnection_count = 0

        # 初始化公共客户端
        self._public_client = self._create_client(OKXWsBaseUrlType.PUBLIC)

        # 只有在提供了完整的API凭证时才创建私有和业务客户端
        self._private_client = None
        self._business_client = None
        if all([api_key, api_secret, passphrase]):
            self._private_client = self._create_client(OKXWsBaseUrlType.PRIVATE)
            self._business_client = self._create_client(OKXWsBaseUrlType.BUSINESS)

    async def _on_reconnect(self) -> None:
        """Handle reconnection events."""
        pass

    @property
    def is_connected(self) -> bool:
        """检查所有相关客户端是否已连接"""
        public_connected = (
            self._public_client
            and hasattr(self._public_client, "_client")
            and self._public_client._client
            and self._public_client._client.is_alive()
        )
        if not self._private_client:
            return public_connected

        private_connected = (
            self._private_client
            and hasattr(self._private_client, "_client")
            and self._private_client._client
            and self._private_client._client.is_alive()
        )
        business_connected = (
            self._business_client
            and hasattr(self._business_client, "_client")
            and self._business_client._client
            and self._business_client._client.is_alive()
        )
        return all([public_connected, private_connected, business_connected])

    async def connect(self) -> None:
        """Connect all available WebSocket clients."""
        if self.is_connected:
            return

        await self._public_client.connect()

        if self._private_client:
            await self._private_client.connect()
        if self._business_client:
            await self._business_client.connect()

    async def disconnect(self) -> None:
        """Disconnect all available WebSocket clients."""
        await self._public_client.disconnect()

        if self._private_client:
            await self._private_client.disconnect()
        if self._business_client:
            await self._business_client.disconnect()

    def _create_client(self, ws_type: OKXWsBaseUrlType) -> OKXWebsocketClient:
        """Create a WebSocket client for specific type."""
        return OKXWebsocketClient(
            clock=self._clock,
            handler=self._handler,
            handler_reconnect=self._on_reconnect,
            api_key=self._api_key if ws_type != OKXWsBaseUrlType.PUBLIC else None,
            api_secret=self._api_secret if ws_type != OKXWsBaseUrlType.PUBLIC else None,
            passphrase=self._passphrase if ws_type != OKXWsBaseUrlType.PUBLIC else None,
            base_url=None,  # Use default URL
            ws_base_url_type=ws_type,
            is_demo=self._is_demo,
            loop=self._loop,
        )

    # Public API methods
    async def subscribe_order_book(self, symbol: str, depth: int = 400) -> None:
        """Subscribe to order book updates."""
        await self.connect()
        await self._public_client.subscribe_order_book(
            instId=symbol,
            depth=depth,
        )

    async def subscribe_trade(self, symbol: str) -> None:
        """Subscribe to trade updates."""
        await self.connect()
        await self._public_client.subscribe_trades(instId=symbol)

    # Business API methods
    # TODO: what is business client?
    async def subscribe_candlesticks(self, symbol: str, interval: str) -> None:
        """Subscribe to candlestick data."""
        await self.connect()
        # Convert interval string to OKXBarSize enum
        bar_size_map = {
            "1m": OKXBarSize.MINUTE_1,
            "3m": OKXBarSize.MINUTE_3,
            "5m": OKXBarSize.MINUTE_5,
            "15m": OKXBarSize.MINUTE_15,
            "30m": OKXBarSize.MINUTE_30,
            "1H": OKXBarSize.HOUR_1,
            "2H": OKXBarSize.HOUR_2,
            "4H": OKXBarSize.HOUR_4,
        }
        bar_size = bar_size_map.get(interval)
        if not bar_size:
            raise ValueError(f"Unsupported interval: {interval}")

        await self._business_client.subscribe_candlesticks(
            instId=symbol,
            bar_size=bar_size,
        )

    # Private API methods
    async def subscribe_account(self) -> None:
        """Subscribe to account updates."""
        await self.connect()
        await self._private_client.subscribe_account()

    async def subscribe_positions(self) -> None:
        """Subscribe to position updates."""
        await self.connect()
        await self._private_client.subscribe_positions()

    async def subscribe_orders(self) -> None:
        """Subscribe to order updates."""
        await self.connect()
        await self._private_client.subscribe_orders()

    async def subscribe_fills(self) -> None:
        """Subscribe to trade fills."""
        await self.connect()
        await self._private_client.subscribe_fills()

    # async def _on_reconnect(self) -> None:
    #     """重连事件处理"""
    #     self.reconnection_count += 1
    #     self.last_reconnect_time = datetime.now()
    #     logger.warning(
    #         f"WebSocket reconnected! Count: {self.reconnection_count}, "
    #         f"Disconnect duration: {self.last_reconnect_time - self.last_disconnect_time}"
    #     )

    # async def simulate_disconnect(self, client_type: str = "public"):
    #     """
    #     模拟断开连接
    #     :param client_type: 'public', 'private', 或 'business'
    #     """
    #     client = getattr(self, f"_{client_type}_client")
    #     if client and hasattr(client, "_client") and client._client:
    #         logger.info(f"Simulating disconnect for {client_type} client...")
    #         self.last_disconnect_time = datetime.now()
    #         await client._client.disconnect()
    #         logger.info(f"{client_type.capitalize()} client disconnected")

    # async def test_reconnection(self, interval: int = 15, times: int = 3):
    #     """
    #     测试重连逻辑
    #     :param interval: 断连间隔(秒)
    #     :param times: 测试次数
    #     """
    #     for i in range(times):
    #         logger.info(f"Starting reconnection test #{i+1}")
    #         await asyncio.sleep(interval)

    #         # 模拟断连
    #         await self.simulate_disconnect("public")
    #         if self._api_key:  # 如果有API key，也测试私有连接
    #             await self.simulate_disconnect("private")
    #             await self.simulate_disconnect("business")

    #         # 等待自动重连
    #         await asyncio.sleep(10)

    #         logger.info(
    #             f"Test #{i+1} completed. "
    #             f"Total reconnections: {self.reconnection_count}"
    #         )



from abc import ABC, abstractmethod
from typing import Dict
from decimal import Decimal


from aiolimiter import AsyncLimiter

from nexustrader.base.ws_client import WSClient
from nexustrader.base.api_client import ApiClient
from nexustrader.schema import Order, BaseMarket
from nexustrader.constants import ExchangeType
from nexustrader.core.log import SpdLog
from nexustrader.core.cache import AsyncCache
from nexustrader.core.entity import RateLimit
from nexustrader.constants import OrderSide, OrderType, TimeInForce, PositionSide
from nexustrader.core.nautilius_core import LiveClock, MessageBus



class PublicConnector(ABC):
    def __init__(
        self,
        account_type,
        market: Dict[str, BaseMarket],
        market_id: Dict[str, str],
        exchange_id: ExchangeType,
        ws_client: WSClient,
        msgbus: MessageBus,
    ):
        self._log = SpdLog.get_logger(
            name=type(self).__name__, level="DEBUG", flush=True
        )
        self._account_type = account_type
        self._market = market
        self._market_id = market_id
        self._exchange_id = exchange_id
        self._ws_client = ws_client
        self._msgbus = msgbus
        self._clock = LiveClock()

    @property
    def account_type(self):
        return self._account_type

    @abstractmethod
    async def subscribe_trade(self, symbol: str):
        """Subscribe to the trade data"""
        pass

    @abstractmethod
    async def subscribe_bookl1(self, symbol: str):
        """Subscribe to the bookl1 data"""
        pass

    @abstractmethod
    async def subscribe_kline(self, symbol: str, interval: str):
        """Subscribe to the kline data"""
        pass

    async def disconnect(self):
        """Disconnect from the exchange"""
        self._ws_client.disconnect()  # not needed to await


class PrivateConnector(ABC):
    def __init__(
        self,
        account_type,
        market: Dict[str, BaseMarket],
        market_id: Dict[str, str],
        exchange_id: ExchangeType,
        ws_client: WSClient,
        api_client: ApiClient,
        msgbus: MessageBus,
        cache: AsyncCache,
        rate_limit: RateLimit | None = None,
    ):
        self._log = SpdLog.get_logger(
            name=type(self).__name__, level="DEBUG", flush=True
        )
        self._account_type = account_type
        self._market = market
        self._market_id = market_id
        self._exchange_id = exchange_id
        self._ws_client = ws_client
        self._api_client = api_client
        self._cache = cache
        self._clock = LiveClock()
        self._msgbus: MessageBus = msgbus

        if rate_limit:
            self._limiter = AsyncLimiter(rate_limit.max_rate, rate_limit.time_period)
        else:
            self._limiter = None

    @property
    def account_type(self):
        return self._account_type

    @abstractmethod
    async def _init_account_balance(self):
        """Initialize the account balance"""
        pass
    
    @abstractmethod
    async def _init_position(self):
        """Initialize the position"""
        pass

    @abstractmethod
    async def create_order(
        self,
        symbol: str,
        side: OrderSide,
        type: OrderType,
        amount: Decimal,
        price: Decimal,
        time_in_force: TimeInForce,
        position_side: PositionSide,
        **kwargs,
    ) -> Order:
        """Create an order"""
        pass

    @abstractmethod
    async def cancel_order(self, symbol: str, order_id: str, **kwargs) -> Order:
        """Cancel an order"""
        pass

    @abstractmethod
    async def connect(self):
        """Connect to the exchange"""
        await self._init_account_balance()
        await self._init_position()

    async def disconnect(self):
        """Disconnect from the exchange"""
        self._ws_client.disconnect()
        await self._api_client.close_session()



from typing import Any, Dict
from nexustrader.base import ExchangeManager
import ccxt
import orjson
import msgspec
from nexustrader.exchange.okx.schema import OkxMarket


class OkxExchangeManager(ExchangeManager):
    api: ccxt.okx
    market: Dict[str, OkxMarket] # symbol -> okx market
    market_id: Dict[str, str] # symbol -> exchange symbol id

    def __init__(self, config: Dict[str, Any] = None):
        config = config or {}
        config["exchange_id"] = config.get("exchange_id", "okx")
        super().__init__(config)
        self.passphrase = config.get("password", None)
    
    def load_markets(self):
        market = self.api.load_markets()
        for symbol, mkt in market.items():
            try:
                mkt_json = orjson.dumps(mkt)
                mkt = msgspec.json.decode(mkt_json, type=OkxMarket)
                
                if (mkt.spot or mkt.linear or mkt.inverse or mkt.future) and not mkt.option:
                    symbol = self._parse_symbol(mkt, exchange_suffix="OKX")
                    mkt.symbol = symbol
                    self.market[symbol] = mkt
                    self.market_id[mkt.id] = symbol # since okx symbol id is identical, no need to distinguish spot, linear, inverse
                
            except Exception as e:
                print(f"Error: {e}, {symbol}, {mkt}")
                continue



from typing import Literal, Callable
from typing import Any
from aiolimiter import AsyncLimiter


from nexustrader.base import WSClient
from nexustrader.exchange.binance.constants import BinanceAccountType
from nexustrader.core.entity import TaskManager


class BinanceWSClient(WSClient):
    def __init__(
        self,
        account_type: BinanceAccountType,
        handler: Callable[..., Any],
        task_manager: TaskManager,
    ):
        self._account_type = account_type
        url = account_type.ws_url
        super().__init__(
            url,
            limiter=AsyncLimiter(max_rate=3, time_period=1),
            handler=handler,
            task_manager=task_manager,
        )

    async def _subscribe(self, params: str, subscription_id: str):
        if subscription_id not in self._subscriptions:
            await self.connect()
            id = self._clock.timestamp_ms()
            payload = {
                "method": "SUBSCRIBE",
                "params": [params],
                "id": id,
            }
            self._subscriptions[subscription_id] = payload
            await self._send(payload)
            self._log.info(f"Subscribing to {subscription_id}...")
        else:
            self._log.info(f"Already subscribed to {subscription_id}")

    async def subscribe_agg_trade(self, symbol: str):
        if (
            self._account_type.is_isolated_margin_or_margin
            or self._account_type.is_portfolio_margin
        ):
            raise ValueError(
                "Not Supported for `Margin Account` or `Portfolio Margin Account`"
            )
        subscription_id = f"agg_trade.{symbol}"
        params = f"{symbol.lower()}@aggTrade"
        await self._subscribe(params, subscription_id)

    async def subscribe_trade(self, symbol: str):
        if (
            self._account_type.is_isolated_margin_or_margin
            or self._account_type.is_portfolio_margin
        ):
            raise ValueError(
                "Not Supported for `Margin Account` or `Portfolio Margin Account`"
            )
        subscription_id = f"trade.{symbol}"
        params = f"{symbol.lower()}@trade"
        await self._subscribe(params, subscription_id)

    async def subscribe_book_ticker(self, symbol: str):
        if (
            self._account_type.is_isolated_margin_or_margin
            or self._account_type.is_portfolio_margin
        ):
            raise ValueError(
                "Not Supported for `Margin Account` or `Portfolio Margin Account`"
            )
        subscription_id = f"book_ticker.{symbol}"
        params = f"{symbol.lower()}@bookTicker"
        await self._subscribe(params, subscription_id)

    async def subscribe_mark_price(
        self, symbol: str, interval: Literal["1s", "3s"] = "1s"
    ):
        if not self._account_type.is_future:
            raise ValueError("Only Supported for `Future Account`")
        subscription_id = f"mark_price.{symbol}"
        params = f"{symbol.lower()}@markPrice@{interval}"
        await self._subscribe(params, subscription_id)

    async def subscribe_user_data_stream(self, listen_key: str):
        subscription_id = "user_data_stream"
        await self._subscribe(listen_key, subscription_id)

    async def subscribe_kline(
        self,
        symbol: str,
        interval: Literal[
            "1s",
            "1m",
            "3m",
            "5m",
            "15m",
            "30m",
            "1h",
            "2h",
            "4h",
            "6h",
            "8h",
            "12h",
            "1d",
            "3d",
            "1w",
            "1M",
        ],
    ):
        if (
            self._account_type.is_isolated_margin_or_margin
            or self._account_type.is_portfolio_margin
        ):
            raise ValueError(
                "Not Supported for `Margin Account` or `Portfolio Margin Account`"
            )
        subscription_id = f"kline.{symbol}.{interval}"
        params = f"{symbol.lower()}@kline_{interval}"
        await self._subscribe(params, subscription_id)

    async def _resubscribe(self):
        for _, payload in self._subscriptions.items():
            await self._send(payload)



import asyncio
from collections import defaultdict
from typing import Optional
from nexustrader.core.log import SpdLog
from nexustrader.schema import Order

class OrderRegistry:
    def __init__(self):
        self._log = SpdLog.get_logger(
            name=type(self).__name__, level="DEBUG", flush=True
        )
        self._uuid_to_order_id = {}
        self._order_id_to_uuid = {}
        self._uuid_init_events = defaultdict(asyncio.Event)

    def register_order(self, order: Order) -> None:
        """Register a new order ID to UUID mapping"""
        self._order_id_to_uuid[order.id] = order.uuid
        self._uuid_to_order_id[order.uuid] = order.id
        self._uuid_init_events[order.id].set() # the order id is linked to the order submit uuid
        self._log.debug(f"[ORDER REGISTER]: linked order id {order.id} with uuid {order.uuid}")

    def get_order_id(self, uuid: str) -> Optional[str]:
        """Get order ID by UUID"""
        return self._uuid_to_order_id.get(uuid, None)

    def get_uuid(self, order_id: str) -> Optional[str]:
        """Get UUID by order ID"""
        return self._order_id_to_uuid.get(order_id, None)

    async def wait_for_order_id(self, order_id: str) -> None:
        """Wait for an order ID to be registered"""
        await self._uuid_init_events[order_id].wait()

    def remove_order(self, order: Order) -> None:
        """Remove order mapping when no longer needed"""
        self._order_id_to_uuid.pop(order.id, None)
        self._uuid_to_order_id.pop(order.uuid, None)
        self._uuid_init_events.pop(order.id, None)



from dataclasses import dataclass
from typing import Dict, List
from nexustrader.constants import AccountType, ExchangeType
from nexustrader.core.entity import RateLimit
from nexustrader.strategy import Strategy
from zmq.asyncio import Socket

@dataclass
class BasicConfig:
    api_key: str
    secret: str
    testnet: bool = False
    passphrase: str = None

@dataclass
class PublicConnectorConfig:
    account_type: AccountType

@dataclass
class PrivateConnectorConfig:
    account_type: AccountType
    rate_limit: RateLimit | None = None
    
@dataclass
class ZeroMQSignalConfig:
    """ZeroMQ Signal Configuration Class.

    Used to configure the ZeroMQ subscriber socket to receive custom trade signals.

    Attributes:
        socket (`zmq.asyncio.Socket`): ZeroMQ asynchronous socket object

    Example:
        >>> from zmq.asyncio import Context
        >>> context = Context()
        >>> socket = context.socket(zmq.SUB)
        >>> socket.connect("ipc:///tmp/zmq_custom_signal")
        >>> socket.setsockopt(zmq.SUBSCRIBE, b"")
        >>> config = ZeroMQSignalConfig(socket=socket)
    """
    socket: Socket
    

@dataclass
class Config:
    strategy_id: str
    user_id: str
    strategy: Strategy
    basic_config: Dict[ExchangeType, BasicConfig]
    public_conn_config: Dict[ExchangeType, List[PublicConnectorConfig]]
    private_conn_config: Dict[ExchangeType, List[PrivateConnectorConfig]]
    zero_mq_signal_config: ZeroMQSignalConfig | None = None
    cache_sync_interval: int = 60
    cache_expire_time: int = 3600



import asyncio
from typing import Dict
from decimal import Decimal
from nexustrader.constants import AccountType
from nexustrader.schema import OrderSubmit
from nexustrader.core.cache import AsyncCache
from nexustrader.core.nautilius_core import MessageBus
from nexustrader.core.entity import TaskManager
from nexustrader.core.registry import OrderRegistry
from nexustrader.exchange.bybit import BybitAccountType
from nexustrader.exchange.bybit.schema import BybitMarket
from nexustrader.base import ExecutionManagementSystem

class BybitExecutionManagementSystem(ExecutionManagementSystem):
    _market: Dict[str, BybitMarket]

    def __init__(
        self,
        market: Dict[str, BybitMarket],
        cache: AsyncCache,
        msgbus: MessageBus,
        task_manager: TaskManager,
        registry: OrderRegistry,
    ):
        super().__init__(
            market=market,
            cache=cache,
            msgbus=msgbus,
            task_manager=task_manager,
            registry=registry,
        )
        self._bybit_account_type: BybitAccountType = None

    def _build_order_submit_queues(self):
        for account_type in self._private_connectors.keys():
            if isinstance(account_type, BybitAccountType):
                self._order_submit_queues[account_type] = asyncio.Queue()

    def _set_account_type(self):
        account_types = self._private_connectors.keys()
        self._bybit_account_type = (
            BybitAccountType.UNIFIED_TESTNET
            if BybitAccountType.UNIFIED_TESTNET in account_types
            else BybitAccountType.UNIFIED
        )

    def _submit_order(
        self, order: OrderSubmit, account_type: AccountType | None = None
    ):
        if not account_type:
            account_type = self._bybit_account_type
        self._order_submit_queues[account_type].put_nowait(order)
        
    def _get_min_order_amount(self, symbol: str, market: BybitMarket) -> Decimal:
        book = self._cache.bookl1(symbol)
        min_order_amount = max(12 / (book.bid + book.ask), market.limits.amount.min)
        min_order_amount = self._amount_to_precision(symbol, min_order_amount, mode="ceil")
        return min_order_amount



import os
import sys
from typing import Literal, Union, Dict, List
from enum import Enum
from dynaconf import Dynaconf


def is_sphinx_build():
    return 'sphinx' in sys.modules

if not os.path.exists(".keys/"):
    os.makedirs(".keys/")
if not os.path.exists(".keys/.secrets.toml") and not is_sphinx_build():
    raise FileNotFoundError(
        "Config file not found, please create a config file at .keys/.secrets.toml"
    )


settings = Dynaconf(
    envvar_prefix="NEXUS",
    settings_files=['.keys/settings.toml', '.keys/.secrets.toml'],
    load_dotenv=True,
)

def get_redis_config(in_docker: bool = False):
    try:
        if in_docker:
            return {
                "host": "redis",
                "db": settings.REDIS_DB,
                "password": settings.REDIS_PASSWORD,
            }

        return {
            "host": settings.REDIS_HOST,
            "port": settings.REDIS_PORT,
            "db": settings.REDIS_DB,
            "password": settings.REDIS_PASSWORD,
        }
    except Exception as e:
        raise ValueError(f"Failed to get Redis password: {e}")


class Url:
    class Bybit:
        class Spot:
            MAINNET = "wss://stream.bybit.com/v5/public/spot"
            TESTNET = "wss://stream-testnet.bybit.com/v5/public/spot"

        class Linear:
            MAINNET = "wss://stream.bybit.com/v5/public/linear"
            TESTNET = "wss://stream-testnet.bybit.com/v5/public/linear"

        class Inverse:
            MAINNET = "wss://stream.bybit.com/v5/public/inverse"
            TESTNET = "wss://stream-testnet.bybit.com/v5/public/inverse"

        class Option:
            MAINNET = "wss://stream.bybit.com/v5/public/option"
            TESTNET = "wss://stream-testnet.bybit.com/v5/public/option"

    class Binance:
        class Spot:
            BASE_URL = "https://api.binance.com/api/v3/userDataStream"
            STREAM_URL = "wss://stream.binance.com:9443/ws"

        class Margin:
            BASE_URL = "https://api.binance.com/sapi/v1/userDataStream"
            STREAM_URL = "wss://stream.binance.com:9443/ws"

        class IsolatedMargin:
            BASE_URL = "https://api.binance.com/sapi/v1/userDataStream/isolated"
            STREAM_URL = "wss://stream.binance.com:9443/ws"

        class UsdMFuture:
            BASE_URL = "https://fapi.binance.com/fapi/v1/listenKey"
            STREAM_URL = "wss://fstream.binance.com/ws"

        class CoinMFuture:
            BASE_URL = "https://dapi.binance.com/dapi/v1/listenKey"
            STREAM_URL = "wss://dstream.binance.com/ws"

        class PortfolioMargin:
            BASE_URL = "https://papi.binance.com/papi/v1/listenKey"
            STREAM_URL = "wss://fstream.binance.com/pm/ws"

        class SpotTestnet:
            BASE_URL = "https://testnet.binance.vision/api/v3/userDataStream"
            STREAM_URL = "wss://testnet.binance.vision/ws"

        class UsdMFutureTestnet:
            BASE_URL = "https://testnet.binancefuture.com/fapi/v1/listenKey"
            STREAM_URL = "wss://stream.binancefuture.com/ws"

        class CoinMFutureTestnet:
            BASE_URL = "https://testnet.binancefuture.com/dapi/v1/listenKey"
            STREAM_URL = "wss://dstream.binancefuture.com/ws"

    class Okx:
        LIVE = "wss://ws.okx.com:8443/ws"
        AWS = "wss://wsaws.okx.com:8443/ws"
        DEMO = "wss://wspap.okx.com:8443/ws"

        # class Live:
        #     PUBLIC = "wss://ws.okx.com:8443/ws/v5/public"
        #     PRIVATE = "wss://ws.okx.com:8443/ws/v5/private"
        #     BUSINESS = "wss://ws.okx.com:8443/ws/v5/business"

        # class Aws:
        #     PUBLIC = "wss://wsaws.okx.com:8443/ws/v5/public"
        #     PRIVATE = "wss://wsaws.okx.com:8443/ws/v5/private"
        #     BUSINESS = "wss://wsaws.okx.com:8443/ws/v5/business"

        # class Demo:
        #     PUBLIC = "wss://wspap.okx.com:8443/ws/v5/public"
        #     PRIVATE = "wss://wspap.okx.com:8443/ws/v5/private"
        #     BUSINESS = "wss://wspap.okx.com:8443/ws/v5/business"


IntervalType = Literal[
    "1s",
    "1m",
    "3m",
    "5m",
    "15m",
    "30m",
    "1h",
    "2h",
    "4h",
    "6h",
    "8h",
    "12h",
    "1d",
    "3d",
    "1w",
    "1M",
]


UrlType = Union[
    Url.Binance.Spot,
    Url.Binance.Margin,
    Url.Binance.IsolatedMargin,
    Url.Binance.UsdMFuture,
    Url.Binance.CoinMFuture,
    Url.Binance.PortfolioMargin,
    Url.Binance.SpotTestnet,
    Url.Binance.UsdMFutureTestnet,
    Url.Binance.CoinMFutureTestnet,
    Url.Okx,
    Url.Bybit.Spot,
    Url.Bybit.Linear,
    Url.Bybit.Inverse,
    Url.Bybit.Option,
]

class SubmitType(Enum):
    CREATE = 0
    CANCEL = 1
    TWAP = 2
    CANCEL_TWAP = 3
    VWAP = 4
    CANCEL_VWAP = 5
class EventType(Enum):
    BOOKL1 = 0
    TRADE = 1
    KLINE = 2
    MARK_PRICE = 3
    FUNDING_RATE = 4
    INDEX_PRICE = 5


class AlgoOrderStatus(Enum):
    RUNNING = "RUNNING"
    CANCELING = "CANCELING"
    FINISHED = "FINISHED"
    CANCELED = "CANCELED"
    FAILED = "FAILED"

class OrderStatus(Enum):
    # LOCAL
    INITIALIZED = "INITIALIZED"
    FAILED = "FAILED"
    CANCEL_FAILED = "CANCEL_FAILED"

    # IN-FLOW
    PENDING = "PENDING"
    CANCELING = "CANCELING"

    # OPEN
    ACCEPTED = "ACCEPTED"
    PARTIALLY_FILLED = "PARTIALLY_FILLED"

    # CLOSED
    FILLED = "FILLED"
    CANCELED = "CANCELED"
    EXPIRED = "EXPIRED"


class ExchangeType(Enum):
    BINANCE = "binance"
    OKX = "okx"
    BYBIT = "bybit"


class BinanceAccountType(Enum):
    SPOT = "SPOT"
    MARGIN = "MARGIN"
    ISOLATED_MARGIN = "ISOLATED_MARGIN"
    USD_M_FUTURE = "USD_M_FUTURE"
    COIN_M_FUTURE = "COIN_M_FUTURE"
    PORTFOLIO_MARGIN = "PORTFOLIO_MARGIN"
    SPOT_TESTNET = "SPOT_TESTNET"
    USD_M_FUTURE_TESTNET = "USD_M_FUTURE_TESTNET"
    COIN_M_FUTURE_TESTNET = "COIN_M_FUTURE_TESTNET"


class OkxAccountType(Enum):
    LIVE = 0
    AWS = 1
    DEMO = 2


class BybitAccountType(Enum):
    SPOT = 0
    LINEAR = 1
    INVERSE = 2
    OPTION = 3
    SPOT_TESTNET = 4
    LINEAR_TESTNET = 5
    INVERSE_TESTNET = 6
    OPTION_TESTNET = 7


STREAM_URLS = {
    BinanceAccountType.SPOT: "wss://stream.binance.com:9443/ws",
    BinanceAccountType.MARGIN: "wss://stream.binance.com:9443/ws",
    BinanceAccountType.ISOLATED_MARGIN: "wss://stream.binance.com:9443/ws",
    BinanceAccountType.USD_M_FUTURE: "wss://fstream.binance.com/ws",
    BinanceAccountType.COIN_M_FUTURE: "wss://dstream.binance.com/ws",
    BinanceAccountType.PORTFOLIO_MARGIN: "wss://fstream.binance.com/pm/ws",
    BinanceAccountType.SPOT_TESTNET: "wss://testnet.binance.vision/ws",
    BinanceAccountType.USD_M_FUTURE_TESTNET: "wss://stream.binancefuture.com/ws",
    BinanceAccountType.COIN_M_FUTURE_TESTNET: "wss://dstream.binancefuture.com/ws",
    OkxAccountType.LIVE: "wss://ws.okx.com:8443/ws",
    OkxAccountType.AWS: "wss://wsaws.okx.com:8443/ws",
    OkxAccountType.DEMO: "wss://wspap.okx.com:8443/ws",
    BybitAccountType.SPOT: "wss://stream.bybit.com/v5/public/spot",
    BybitAccountType.LINEAR: "wss://stream.bybit.com/v5/public/linear",
    BybitAccountType.INVERSE: "wss://stream.bybit.com/v5/public/inverse",
    BybitAccountType.OPTION: "wss://stream.bybit.com/v5/public/option",
    BybitAccountType.SPOT_TESTNET: "wss://stream-testnet.bybit.com/v5/public/spot",
    BybitAccountType.LINEAR_TESTNET: "wss://stream-testnet.bybit.com/v5/public/linear",
    BybitAccountType.INVERSE_TESTNET: "wss://stream-testnet.bybit.com/v5/public/inverse",
    BybitAccountType.OPTION_TESTNET: "wss://stream-testnet.bybit.com/v5/public/option",
}

LISTEN_KEY_URLS = {
    BinanceAccountType.SPOT: "https://api.binance.com/api/v3/userDataStream",
    BinanceAccountType.MARGIN: "https://api.binance.com/sapi/v1/userDataStream",
    BinanceAccountType.ISOLATED_MARGIN: "https://api.binance.com/sapi/v1/userDataStream/isolated",
    BinanceAccountType.USD_M_FUTURE: "https://fapi.binance.com/fapi/v1/listenKey",
    BinanceAccountType.COIN_M_FUTURE: "https://dapi.binance.com/dapi/v1/listenKey",
    BinanceAccountType.PORTFOLIO_MARGIN: "https://papi.binance.com/papi/v1/listenKey",
    BinanceAccountType.SPOT_TESTNET: "https://testnet.binance.vision/api/v3/userDataStream",
    BinanceAccountType.USD_M_FUTURE_TESTNET: "https://testnet.binancefuture.com/fapi/v1/listenKey",
    BinanceAccountType.COIN_M_FUTURE_TESTNET: "https://testnet.binancefuture.com/dapi/v1/listenKey",
}

BASE_URLS = {
    BinanceAccountType.SPOT: "https://api.binance.com",
    BinanceAccountType.MARGIN: "https://api.binance.com",
    BinanceAccountType.ISOLATED_MARGIN: "https://api.binance.com",
    BinanceAccountType.USD_M_FUTURE: "https://fapi.binance.com",
    BinanceAccountType.COIN_M_FUTURE: "https://dapi.binance.com",
    BinanceAccountType.PORTFOLIO_MARGIN: "https://papi.binance.com",
    BinanceAccountType.SPOT_TESTNET: "https://testnet.binance.vision",
    BinanceAccountType.USD_M_FUTURE_TESTNET: "https://testnet.binancefuture.com",
    BinanceAccountType.COIN_M_FUTURE_TESTNET: "https://testnet.binancefuture.com",
}


class BinanceEndpointsType(Enum):
    USER_DATA_STREAM = 0
    ACCOUNT = 1
    TRADING = 2
    MARKET = 3
    GENERAL = 4


BINANCE_ENDPOINTS = {
    BinanceEndpointsType.USER_DATA_STREAM: {
        BinanceAccountType.SPOT: "/api/v3/userDataStream",
        BinanceAccountType.MARGIN: "/sapi/v1/userDataStream",
        BinanceAccountType.ISOLATED_MARGIN: "/sapi/v1/userDataStream/isolated",
        BinanceAccountType.USD_M_FUTURE: "/fapi/v1/listenKey",
        BinanceAccountType.COIN_M_FUTURE: "/dapi/v1/listenKey",
        BinanceAccountType.PORTFOLIO_MARGIN: "/papi/v1/listenKey",
        BinanceAccountType.SPOT_TESTNET: "/api/v3/userDataStream",
        BinanceAccountType.USD_M_FUTURE_TESTNET: "/fapi/v1/listenKey",
        BinanceAccountType.COIN_M_FUTURE_TESTNET: "/dapi/v1/listenKey",
    },
    BinanceEndpointsType.TRADING: {
        BinanceAccountType.SPOT: "/api/v3",
        BinanceAccountType.MARGIN: "/sapi/v1",
        BinanceAccountType.ISOLATED_MARGIN: "/sapi/v1",
        BinanceAccountType.USD_M_FUTURE: "/fapi/v1",
        BinanceAccountType.COIN_M_FUTURE: "/dapi/v1",
        BinanceAccountType.PORTFOLIO_MARGIN: "/papi/v1",
        BinanceAccountType.SPOT_TESTNET: "/api/v3",
        BinanceAccountType.USD_M_FUTURE_TESTNET: "/fapi/v1",
        BinanceAccountType.COIN_M_FUTURE_TESTNET: "/dapi/v1",
    },
}


class WSType(Enum):
    BINANCE_SPOT = 0
    BINANCE_MARGIN = 1
    BINANCE_ISOLATED_MARGIN = 2
    BINANCE_USD_M_FUTURE = 3
    BINANCE_COIN_M_FUTURE = 4
    BINANCE_PORTFOLIO_MARGIN = 5
    BINANCE_SPOT_TESTNET = 6
    BINANCE_USD_M_FUTURE_TESTNET = 7
    OKX_LIVE = 8
    OKX_AWS = 9
    OKX_DEMO = 10


class PublicConnectorType(Enum):
    BINANCE_SPOT = 0
    BINANCE_USD_M_FUTURE = 1
    BINANCE_COIN_M_FUTURE = 2
    BINANCE_SPOT_TESTNET = 3
    BINANCE_USD_M_FUTURE_TESTNET = 4
    BINANCE_COIN_M_FUTURE_TESTNET = 5
    OKX_LIVE = 6
    OKX_AWS = 7
    OKX_DEMO = 8


class AccountType(Enum):
    pass


class OrderType(Enum):
    MARKET = "MARKET"
    LIMIT = "LIMIT"


class OrderSide(Enum):
    BUY = "BUY"
    SELL = "SELL"


class TimeInForce(Enum):
    GTC = "GTC"
    IOC = "IOC"
    FOK = "FOK"


class PositionSide(Enum):
    LONG = "LONG"
    SHORT = "SHORT"
    FLAT = "FLAT"


class InstrumentType(Enum):
    SPOT = "spot"
    MARGIN = "margin"
    FUTURE = "future"
    OPTION = "option"
    SWAP = "swap"
    LINEAR = "linear"
    INVERSE = "inverse"


class OptionType(Enum):
    CALL = "call"
    PUT = "put"

STATUS_TRANSITIONS: Dict[OrderStatus, List[OrderStatus]] = {
    OrderStatus.PENDING: [
        OrderStatus.CANCELED,
        OrderStatus.CANCELING,
        OrderStatus.ACCEPTED,
        OrderStatus.PARTIALLY_FILLED,
        OrderStatus.CANCELED,
        OrderStatus.FILLED,
        OrderStatus.CANCEL_FAILED,
    ],
    OrderStatus.CANCELING: [
        OrderStatus.CANCELED,
        OrderStatus.PARTIALLY_FILLED,
        OrderStatus.FILLED,
    ],
    OrderStatus.ACCEPTED: [
        OrderStatus.PARTIALLY_FILLED,
        OrderStatus.FILLED,
        OrderStatus.CANCELING,
        OrderStatus.CANCELED,
        OrderStatus.EXPIRED,
        OrderStatus.CANCEL_FAILED,
    ],
    OrderStatus.PARTIALLY_FILLED: [
        OrderStatus.PARTIALLY_FILLED,
        OrderStatus.FILLED,
        OrderStatus.CANCELING,
        OrderStatus.CANCELED,
        OrderStatus.EXPIRED,
        OrderStatus.CANCEL_FAILED,
    ],
    OrderStatus.FILLED: [],
    OrderStatus.CANCELED: [],
    OrderStatus.EXPIRED: [],
    OrderStatus.FAILED: [],
}


class DataType(Enum):
    BOOKL1 = "bookl1"
    BOOKL2 = "bookl2"
    TRADE = "trade"
    KLINE = "kline"
    MARK_PRICE = "mark_price"
    FUNDING_RATE = "funding_rate"
    INDEX_PRICE = "index_price"

class StorageBackend(Enum):
    REDIS = "redis"
    SQLITE = "sqlite"



import signal
import asyncio
import socket
from typing import Callable
from typing import Dict, List
import warnings

import redis
import time

from dataclasses import dataclass
from nexustrader.constants import get_redis_config
from nexustrader.core.log import SpdLog
from nexustrader.core.nautilius_core import LiveClock


@dataclass
class RateLimit:
    """
    max_rate: Allow up to max_rate / time_period acquisitions before blocking.
    time_period: Time period in seconds.
    """

    max_rate: float
    time_period: float = 60


class TaskManager:
    def __init__(
        self, loop: asyncio.AbstractEventLoop, enable_signal_handlers: bool = True
    ):
        self._log = SpdLog.get_logger(type(self).__name__, level="DEBUG", flush=True)
        self._tasks: Dict[str, asyncio.Task] = {}
        self._shutdown_event = asyncio.Event()
        self._loop = loop
        if enable_signal_handlers:
            self._setup_signal_handlers()

    def _setup_signal_handlers(self):
        try:
            for sig in (signal.SIGINT, signal.SIGTERM):
                self._loop.add_signal_handler(
                    sig, lambda: self.create_task(self._shutdown())
                )
        except NotImplementedError:
            self._log.warning("Signal handlers not supported on this platform")

    async def _shutdown(self):
        self._shutdown_event.set()
        self._log.debug("Shutdown signal received, cleaning up...")

    def create_task(self, coro: asyncio.coroutines, name: str = None) -> asyncio.Task:
        task = asyncio.create_task(coro, name=name)
        self._tasks[task.get_name()] = task
        task.add_done_callback(self._handle_task_done)
        return task

    def cancel_task(self, name: str) -> bool:
        if name in self._tasks:
            self._tasks[name].cancel()
            return True
        return False

    def _handle_task_done(self, task: asyncio.Task):
        try:
            name = task.get_name()
            self._tasks.pop(name, None)
            task.result()
        except asyncio.CancelledError:
            pass
        except Exception as e:
            self._log.error(f"Error during task done: {e}")
            raise

    async def wait(self):
        try:
            if self._tasks:
                await self._shutdown_event.wait()
                self._log.debug("Shutdown Completed")
        except Exception as e:
            self._log.error(f"Error during wait: {e}")
            raise

    async def cancel(self):
        try:
            for task in self._tasks.values():
                if not task.done():
                    task.cancel()

            if self._tasks:
                results = await asyncio.gather(
                    *self._tasks.values(), return_exceptions=True
                )

                for result in results:
                    if isinstance(result, Exception) and not isinstance(
                        result, asyncio.CancelledError
                    ):
                        self._log.error(f"Task failed during cancellation: {result}")

        except Exception as e:
            self._log.error(f"Error during cancellation: {e}")
            raise
        finally:
            self._tasks.clear()

class RedisClient:
    _params = None

    @classmethod
    def _is_in_docker(cls) -> bool:
        try:
            socket.gethostbyname("redis")
            return True
        except socket.gaierror:
            return False

    @classmethod
    def _get_params(cls) -> dict:
        if cls._params is None:
            in_docker = cls._is_in_docker()
            cls._params = get_redis_config(in_docker)
        return cls._params

    @classmethod
    def get_client(cls) -> redis.Redis:
        return redis.Redis(**cls._get_params())

    @classmethod
    def get_async_client(cls) -> redis.asyncio.Redis:
        return redis.asyncio.Redis(**cls._get_params())


class Clock:
    def __init__(self, tick_size: float = 1.0):
        """
        :param tick_size_s: Time interval of each tick in seconds (supports sub-second precision).
        """
        self._tick_size = tick_size  # Tick size in seconds
        self._current_tick = (time.time() // self._tick_size) * self._tick_size
        self._clock = LiveClock()
        self._tick_callbacks: List[Callable[[float], None]] = []
        self._started = False

    @property
    def tick_size(self) -> float:
        return self._tick_size

    @property
    def current_timestamp(self) -> float:
        return self._clock.timestamp()

    def add_tick_callback(self, callback: Callable[[float], None]):
        """
        Register a callback to be called on each tick.
        :param callback: Function to be called with current_tick as argument.
        """
        self._tick_callbacks.append(callback)

    async def run(self):
        if self._started:
            raise RuntimeError("Clock is already running.")
        self._started = True
        while True:
            now = time.time()
            next_tick_time = self._current_tick + self._tick_size
            sleep_duration = next_tick_time - now
            if sleep_duration > 0:
                await asyncio.sleep(sleep_duration)
            else:
                # If we're behind schedule, skip to the next tick to prevent drift
                next_tick_time = now
            self._current_tick = next_tick_time
            for callback in self._tick_callbacks:
                if asyncio.iscoroutinefunction(callback):
                    await callback(self.current_timestamp)
                else:
                    callback(self.current_timestamp)


class ZeroMQSignalRecv:
    def __init__(self, config, callback: Callable, task_manager: TaskManager):
        self._socket = config.socket
        self._callback = callback
        self._task_manager = task_manager

    async def _recv(self):
        while True:
            date = await self._socket.recv()
            if asyncio.iscoroutinefunction(self._callback):
                await self._callback(date)
            else:
                self._callback(date)

    async def start(self):
        self._task_manager.create_task(self._recv())


class DataReady:
    def __init__(self, symbols: List[str], timeout: int = 60):
        """
        Initialize DataReady class

        Args:
            symbols: symbols list need to monitor
            timeout: timeout in seconds
        """
        self._log = SpdLog.get_logger(type(self).__name__, level="DEBUG", flush=True)
        self._symbols = {symbol: False for symbol in symbols}
        self._timeout = timeout
        self._clock = LiveClock()
        self._first_data_time: int | None = None

    def input(self, data) -> None:
        """
        Input data, update the status of the corresponding symbol

        Args:
            data: data object with symbol attribute
        """

        if not self.ready:
            symbol = data.symbol

            if self._first_data_time is None:
                self._first_data_time = self._clock.timestamp_ms()

            if symbol in self._symbols:
                self._symbols[symbol] = True

    @property
    def ready(self) -> bool:
        """
        Check if all data is ready or if it has timed out

        Returns:
            bool: if all data is ready or timed out, return True
        """
        if self._first_data_time is None:
            return False

        if self._clock.timestamp_ms() - self._first_data_time > self._timeout * 1000:
            not_ready = [symbol for symbol, ready in self._symbols.items() if not ready]
            if not_ready:
                warnings.warn(
                    f"Data receiving timed out. The following symbols are not ready: {', '.join(not_ready)}"
                )
            return True

        # check if all data is ready
        return all(self._symbols.values())



import asyncio
from abc import ABC

from nexustrader.schema import Order
from nexustrader.core.log import SpdLog
from nexustrader.core.entity import TaskManager
from nexustrader.core.nautilius_core import MessageBus
from nexustrader.core.cache import AsyncCache
from nexustrader.core.registry import OrderRegistry
from nexustrader.constants import OrderStatus



class OrderManagementSystem(ABC):
    def __init__(
        self,
        cache: AsyncCache,
        msgbus: MessageBus,
        task_manager: TaskManager,
        registry: OrderRegistry,
    ):
        self._log = SpdLog.get_logger(
            name=type(self).__name__, level="DEBUG", flush=True
        )
        self._cache = cache
        self._msgbus = msgbus
        self._task_manager = task_manager
        self._registry = registry

        self._order_msg_queue: asyncio.Queue[Order] = asyncio.Queue()

    def _add_order_msg(self, order: Order):
        """
        Add an order to the order message queue
        """
        self._order_msg_queue.put_nowait(order)

    async def _handle_order_event(self):
        """
        Handle the order event
        """
        while True:
            try:
                order = await self._order_msg_queue.get()

                # handle the ACCEPTED, PARTIALLY_FILLED, CANCELED, FILLED, EXPIRED arived early than the order submit uuid
                uuid = self._registry.get_uuid(order.id)
                if not uuid:
                    await self._registry.wait_for_order_id(order.id)
                    uuid = self._registry.get_uuid(order.id)
                order.uuid = uuid

                match order.status:
                    case OrderStatus.ACCEPTED:
                        self._log.debug(f"ORDER STATUS ACCEPTED: {str(order)}")
                        self._cache._order_status_update(order)
                        self._msgbus.send(endpoint="accepted", msg=order)
                    case OrderStatus.PARTIALLY_FILLED:
                        self._log.debug(f"ORDER STATUS PARTIALLY FILLED: {str(order)}")
                        self._cache._order_status_update(order)
                        self._msgbus.send(endpoint="partially_filled", msg=order)
                    case OrderStatus.CANCELED:
                        self._log.debug(f"ORDER STATUS CANCELED: {str(order)}")
                        self._cache._order_status_update(order)
                        self._msgbus.send(endpoint="canceled", msg=order)
                        self._registry.remove_order(order)
                    case OrderStatus.FILLED:
                        self._log.debug(f"ORDER STATUS FILLED: {str(order)}")
                        self._cache._order_status_update(order)
                        self._msgbus.send(endpoint="filled", msg=order)
                        self._registry.remove_order(order)
                    case OrderStatus.EXPIRED:
                        self._log.debug(f"ORDER STATUS EXPIRED: {str(order)}")
                        self._cache._order_status_update(order)
                self._order_msg_queue.task_done()
            except Exception as e:
                self._log.error(f"Error in handle_order_event: {e}")

    async def start(self):
        """
        Start the order management system
        """
        self._log.debug("OrderManagementSystem started")

        # Start order and position event handlers
        self._task_manager.create_task(self._handle_order_event())



import msgspec
from typing import Dict, Any
import orjson
import hmac
import base64
import asyncio
import aiohttp
from urllib.parse import urljoin, urlencode
from nexustrader.base import ApiClient
from nexustrader.exchange.okx.constants import OkxRestUrl
from nexustrader.exchange.okx.error import OkxHttpError, OkxRequestError
from nexustrader.exchange.okx.schema import (
    OkxPlaceOrderResponse,
    OkxCancelOrderResponse,
    OkxGeneralResponse,
    OkxErrorResponse,
    OkxBalanceResponse,
    OkxPositionResponse,
)
from nexustrader.core.nautilius_core import hmac_signature


class OkxApiClient(ApiClient):
    def __init__(
        self,
        api_key: str = None,
        secret: str = None,
        passphrase: str = None,
        testnet: bool = False,
        timeout: int = 10,
    ):
        super().__init__(
            api_key=api_key,
            secret=secret,
            timeout=timeout,
        )

        self._base_url = OkxRestUrl.DEMO.value if testnet else OkxRestUrl.LIVE.value
        self._passphrase = passphrase
        self._testnet = testnet
        self._place_order_decoder = msgspec.json.Decoder(OkxPlaceOrderResponse)
        self._cancel_order_decoder = msgspec.json.Decoder(OkxCancelOrderResponse)
        self._general_response_decoder = msgspec.json.Decoder(OkxGeneralResponse)
        self._error_response_decoder = msgspec.json.Decoder(OkxErrorResponse)
        self._balance_response_decoder = msgspec.json.Decoder(OkxBalanceResponse, strict=False)
        self._position_response_decoder = msgspec.json.Decoder(OkxPositionResponse, strict=False)
        self._headers = {
            "Content-Type": "application/json",
            "User-Agent": "TradingBot/1.0",
        }
        
    async def get_api_v5_account_balance(self, ccy: str | None = None) -> OkxBalanceResponse:
        endpoint = "/api/v5/account/balance"
        payload = {"ccy": ccy} if ccy else {}
        raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
        return self._balance_response_decoder.decode(raw)
    
    async def get_api_v5_account_positions(self, inst_type: str | None = None, inst_id: str | None = None, pos_id: str | None = None) -> OkxPositionResponse:
        endpoint = "/api/v5/account/positions"
        payload = {
            k: v for k, v in {
                "instType": inst_type,
                "instId": inst_id,
                "posId": pos_id
            }.items() if v is not None
        }
        raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
        return self._position_response_decoder.decode(raw)
    
    async def post_api_v5_trade_order(
        self,
        inst_id: str,
        td_mode: str,
        side: str,
        ord_type: str,
        sz: str,
        **kwargs,
    ) -> OkxPlaceOrderResponse:
        """
        Place a new order
        https://www.okx.com/docs-v5/en/#order-book-trading-trade-post-place-order

        {'arg': {'channel': 'orders', 'instType': 'ANY', 'uid': '611800569950521616'}, 'data': [{'instType': 'SWAP', 'instId': 'BTC-USDT-SWAP', 'tgtCcy': '', 'ccy': '', 'ordId': '1993784914940116992', 'clOrdId': '', 'algoClOrdId': '', 'algoId': '', 'tag': '', 'px': '80000', 'sz': '0.1', 'notionalUsd': '80.0128', 'ordType': 'limit', 'side': 'buy', 'posSide': 'long', 'tdMode': 'cross', 'accFillSz': '0', 'fillNotionalUsd': '', 'avgPx': '0', 'state': 'canceled', 'lever': '3', 'pnl': '0', 'feeCcy': 'USDT', 'fee': '0', 'rebateCcy': 'USDT', 'rebate': '0', 'category': 'normal', 'uTime': '1731921825881', 'cTime': '1731921820806', 'source': '', 'reduceOnly': 'false', 'cancelSource': '1', 'quickMgnType': '', 'stpId': '', 'stpMode': 'cancel_maker', 'attachAlgoClOrdId': '', 'lastPx': '91880', 'isTpLimit': 'false', 'slTriggerPx': '', 'slTriggerPxType': '', 'tpOrdPx': '', 'tpTriggerPx': '', 'tpTriggerPxType': '', 'slOrdPx': '', 'fillPx': '', 'tradeId': '', 'fillSz': '0', 'fillTime': '', 'fillPnl': '0', 'fillFee': '0', 'fillFeeCcy': '', 'execType': '', 'fillPxVol': '', 'fillPxUsd': '', 'fillMarkVol': '', 'fillFwdPx': '', 'fillMarkPx': '', 'amendSource': '', 'reqId': '', 'amendResult': '', 'code': '0', 'msg': '', 'pxType': '', 'pxUsd': '', 'pxVol': '', 'linkedAlgoOrd': {'algoId': ''}, 'attachAlgoOrds': []}]}
        """
        endpoint = "/api/v5/trade/order"
        payload = {
            "instId": inst_id,
            "tdMode": td_mode,
            "side": side,
            "ordType": ord_type,
            "sz": sz,
            **kwargs,
        }
        raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
        return self._place_order_decoder.decode(raw)

    async def post_api_v5_trade_cancel_order(
        self, inst_id: str, ord_id: str | None = None, cl_ord_id: str | None = None
    ) -> OkxCancelOrderResponse:
        """
        Cancel an existing order
        https://www.okx.com/docs-v5/en/#order-book-trading-trade-post-cancel-order
        """
        endpoint = "/api/v5/trade/cancel-order"
        payload = {"instId": inst_id}
        if ord_id:
            payload["ordId"] = ord_id
        if cl_ord_id:
            payload["clOrdId"] = cl_ord_id

        raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
        return self._cancel_order_decoder.decode(raw)

    def _generate_signature(self, message: str) -> str:
        mac = hmac.new(
            bytes(self._secret, encoding="utf8"),
            bytes(message, encoding="utf-8"),
            digestmod="sha256",
        )
        return base64.b64encode(mac.digest()).decode()

    def _generate_signature_v2(self, message: str) -> str:
        hex_digest = hmac_signature(self._secret, message)
        digest = bytes.fromhex(hex_digest)
        return base64.b64encode(digest).decode()

    async def _get_signature(
        self, ts: str, method: str, request_path: str, payload: Dict[str, Any] = None
    ) -> str:
        body = ""
        if payload:
            body = orjson.dumps(payload).decode()

        sign_str = f"{ts}{method}{request_path}{body}"
        signature = self._generate_signature_v2(sign_str)
        return signature

    def _get_timestamp(self) -> str:
        return (
            self._clock.utc_now()
            .isoformat(timespec="milliseconds")
            .replace("+00:00", "Z")
        )

    async def _get_headers(
        self, ts: str, method: str, request_path: str, payload: Dict[str, Any] = None
    ) -> Dict[str, Any]:
        headers = self._headers
        signature = await self._get_signature(ts, method, request_path, payload)
        headers.update(
            {
                "OK-ACCESS-KEY": self._api_key,
                "OK-ACCESS-SIGN": signature,
                "OK-ACCESS-TIMESTAMP": ts,
                "OK-ACCESS-PASSPHRASE": self._passphrase,
            }
        )
        if self._testnet:
            headers["x-simulated-trading"] = "1"
        return headers

    async def _fetch(
        self,
        method: str,
        endpoint: str,
        payload: Dict[str, Any] = None,
        signed: bool = False,
    ) -> bytes:
        self._init_session()
        url = urljoin(self._base_url, endpoint)
        request_path = endpoint
        headers = self._headers
        timestamp = self._get_timestamp()

        payload = payload or {}

        payload_json = urlencode(payload) if method == "GET" else orjson.dumps(payload)

        if method == "GET":
            url += f"?{payload_json}"
            payload_json = None

        if signed and self._api_key:
            headers = await self._get_headers(timestamp, method, request_path, payload)

        try:
            self._log.debug(
                f"Request {method} Url: {url} Headers: {headers} Payload: {payload_json}"
            )

            response = await self._session.request(
                method=method,
                url=url,
                headers=headers,
                data=payload_json,
            )
            raw = await response.read()

            if response.status >= 400:
                raise OkxHttpError(
                    status_code=response.status,
                    message=orjson.loads(raw),
                    headers=response.headers,
                )
            okx_response = self._general_response_decoder.decode(raw)
            if okx_response.code == "0":
                return raw
            else:
                okx_error_response = self._error_response_decoder.decode(raw)
                for data in okx_error_response.data:
                    raise OkxRequestError(
                        error_code=data.sCode,
                        status_code=response.status,
                        message=data.sMsg,
                    )
        except aiohttp.ClientError as e:
            self._log.error(f"Client Error {method} Url: {url} {e}")
            raise
        except asyncio.TimeoutError:
            self._log.error(f"Timeout {method} Url: {url}")
            raise
        except Exception as e:
            self._log.error(f"Error {method} Url: {url} {e}")
            raise



import msgspec
import asyncio
import aiosqlite
import sqlite3

from typing import Dict, Set, Type, List, Optional
from collections import defaultdict
from returns.maybe import maybe
from pathlib import Path

from nexustrader.schema import (
    Order,
    Position,
    ExchangeType,
    InstrumentId,
    Kline,
    BookL1,
    Trade,
    AlgoOrder,
    AccountBalance,
    Balance,
)
from nexustrader.constants import STATUS_TRANSITIONS, AccountType
from nexustrader.core.entity import TaskManager, RedisClient
from nexustrader.core.log import SpdLog
from nexustrader.core.nautilius_core import LiveClock, MessageBus
from nexustrader.constants import StorageBackend

class AsyncCache:
    def __init__(
        self,
        strategy_id: str,
        user_id: str,
        msgbus: MessageBus,
        task_manager: TaskManager,
        storage_backend: StorageBackend = StorageBackend.REDIS,
        db_path: str = "cache.db",
        sync_interval: int = 60,  # seconds
        expire_time: int = 3600,  # seconds
    ):
        self.strategy_id = strategy_id
        self.user_id = user_id
        self._storage_backend = storage_backend
        self._db_path = db_path

        self._log = SpdLog.get_logger(
            name=type(self).__name__, level="DEBUG", flush=True
        )
        self._clock = LiveClock()

        # in-memory save
        self._mem_closed_orders: Dict[str, bool] = {}  # uuid -> bool
        self._mem_orders: Dict[str, Order] = {}  # uuid -> Order
        self._mem_algo_orders: Dict[str, AlgoOrder] = {}  # uuid -> AlgoOrder
        self._mem_open_orders: Dict[ExchangeType, Set[str]] = defaultdict(set)  # exchange_id -> set(uuid)
        self._mem_symbol_open_orders: Dict[str, Set[str]] = defaultdict(set)  # symbol -> set(uuid)
        self._mem_symbol_orders: Dict[str, Set[str]] = defaultdict(set)  # symbol -> set(uuid)
        self._mem_positions: Dict[str, Position] = {}  # symbol -> Position
        self._mem_account_balance: Dict[AccountType, AccountBalance] = defaultdict(AccountBalance)

        # set params
        self._sync_interval = sync_interval  # sync interval
        self._expire_time = expire_time  # expire time
        self._shutdown_event = asyncio.Event()
        self._task_manager = task_manager

        self._kline_cache: Dict[str, Kline] = {}
        self._bookl1_cache: Dict[str, BookL1] = {}
        self._trade_cache: Dict[str, Trade] = {}

        self._msgbus = msgbus
        self._msgbus.subscribe(topic="kline", handler=self._update_kline_cache)
        self._msgbus.subscribe(topic="bookl1", handler=self._update_bookl1_cache)
        self._msgbus.subscribe(topic="trade", handler=self._update_trade_cache)

    ################# # base functions ####################

    def _encode(self, obj: Order | Position | AlgoOrder) -> bytes:
        return msgspec.json.encode(obj)

    def _decode(
        self, data: bytes, obj_type: Type[Order | Position | AlgoOrder]
    ) -> Order | Position | AlgoOrder:
        return msgspec.json.decode(data, type=obj_type)
    
    async def _init_storage(self):
        """Initialize the storage backend"""
        if self._storage_backend == StorageBackend.REDIS:
            self._r_async = RedisClient.get_async_client()
            self._r = RedisClient.get_client()
        else:
            db_path = Path(self._db_path)
            db_path.parent.mkdir(parents=True, exist_ok=True)
            self._db_async = await aiosqlite.connect(str(db_path))
            self._db = sqlite3.connect(str(db_path))
            await self._init_sqlite_tables()

    async def _init_sqlite_tables(self):
        """Initialize the SQLite tables"""
        async with self._db_async.cursor() as cursor:
            await cursor.executescript("""
                CREATE TABLE IF NOT EXISTS orders (
                    uuid TEXT PRIMARY KEY,
                    strategy_id TEXT,
                    user_id TEXT,
                    symbol TEXT,
                    data BLOB,
                    timestamp INTEGER
                );
                
                CREATE INDEX IF NOT EXISTS idx_orders_symbol 
                ON orders(strategy_id, user_id, symbol);
                
                CREATE TABLE IF NOT EXISTS algo_orders (
                    uuid TEXT PRIMARY KEY,
                    strategy_id TEXT,
                    user_id TEXT,
                    symbol TEXT,
                    data BLOB,
                    timestamp INTEGER
                );
                
                CREATE INDEX IF NOT EXISTS idx_algo_orders_symbol 
                ON algo_orders(strategy_id, user_id, symbol);
                
                CREATE TABLE IF NOT EXISTS positions (
                    symbol TEXT,
                    strategy_id TEXT,
                    user_id TEXT,
                    exchange TEXT,
                    data BLOB,
                    PRIMARY KEY (strategy_id, user_id, symbol)
                );
                
                CREATE TABLE IF NOT EXISTS open_orders (
                    uuid PRIMARY KEY,
                    exchange TEXT,
                    symbol TEXT,
                    strategy_id TEXT,
                    user_id TEXT,
                );
                
                CREATE INDEX IF NOT EXISTS idx_open_orders_uuid 
                ON open_orders(strategy_id, user_id, uuid);
            """)
            await self._db_async.commit()

    async def start(self):
        """Start the cache"""
        await self._init_storage()
        self._task_manager.create_task(self._periodic_sync())

    async def _periodic_sync(self):
        """Periodically sync the cache"""
        while not self._shutdown_event.is_set():
            if self._storage_backend == StorageBackend.REDIS:
                await self._sync_to_redis()
            elif self._storage_backend == StorageBackend.SQLITE:
                await self._sync_to_sqlite()
            self._cleanup_expired_data()
            await asyncio.sleep(self._sync_interval)

    async def _sync_to_redis(self):
        """Sync the cache to Redis"""
        self._log.debug("syncing to redis")
        for uuid, order in self._mem_orders.copy().items():
            orders_key = f"strategy:{self.strategy_id}:user_id:{self.user_id}:orders"
            await self._r_async.hset(orders_key, uuid, self._encode(order))
        
        for uuid, algo_order in self._mem_algo_orders.copy().items():
            algo_orders_key = f"strategy:{self.strategy_id}:user_id:{self.user_id}:algo_orders"
            await self._r_async.hset(algo_orders_key, uuid, self._encode(algo_order))

        for exchange, open_order_uuids in self._mem_open_orders.copy().items():
            open_orders_key = f"strategy:{self.strategy_id}:user_id:{self.user_id}:exchange:{exchange.value}:open_orders"

            await self._r_async.delete(open_orders_key)

            if open_order_uuids:
                await self._r_async.sadd(open_orders_key, *open_order_uuids)

        for symbol, uuids in self._mem_symbol_orders.copy().items():
            instrument_id = InstrumentId.from_str(symbol)
            key = f"strategy:{self.strategy_id}:user_id:{self.user_id}:exchange:{instrument_id.exchange.value}:symbol_orders:{symbol}"
            await self._r_async.delete(key)
            if uuids:
                await self._r_async.sadd(key, *uuids)

        for symbol, uuids in self._mem_symbol_open_orders.copy().items():
            instrument_id = InstrumentId.from_str(symbol)
            key = f"strategy:{self.strategy_id}:user_id:{self.user_id}:exchange:{instrument_id.exchange.value}:symbol_open_orders:{symbol}"
            await self._r_async.delete(key)
            if uuids:
                await self._r_async.sadd(key, *uuids)

        # Add position sync        
        for symbol, position in self._mem_positions.copy().items():
            key = f"strategy:{self.strategy_id}:user_id:{self.user_id}:exchange:{position.exchange.value}:symbol_positions:{symbol}"
            await self._r_async.set(key, self._encode(position))
    
    async def _sync_to_sqlite(self):
        """Sync the cache to SQLite"""
        async with self._db_async.cursor() as cursor:
            # sync orders
            for uuid, order in self._mem_orders.copy().items():
                await cursor.execute(
                    "INSERT OR REPLACE INTO orders (uuid, strategy_id, user_id, symbol, data, timestamp) VALUES (?, ?, ?, ?, ?, ?)",
                    (uuid, self.strategy_id, self.user_id, order.symbol, self._encode(order), order.timestamp)
                )

            # sync algo orders
            for uuid, algo_order in self._mem_algo_orders.copy().items():
                await cursor.execute(
                    "INSERT OR REPLACE INTO algo_orders (uuid, strategy_id, user_id, symbol, data, timestamp) VALUES (?, ?, ?, ?, ?, ?)",
                    (uuid, self.strategy_id, self.user_id, algo_order.symbol, self._encode(algo_order), algo_order.timestamp)
                )

            # sync positions
            for symbol, position in self._mem_positions.copy().items():
                await cursor.execute(
                    "INSERT OR REPLACE INTO positions (symbol, strategy_id, user_id, exchange, data) VALUES (?, ?, ?, ?, ?)",
                    (symbol, self.strategy_id, self.user_id, position.exchange.value, self._encode(position))
                )

            # sync open orders
            await cursor.execute("DELETE FROM open_orders WHERE strategy_id = ? AND user_id = ?",
                               (self.strategy_id, self.user_id))
            
            for exchange, uuids in self._mem_open_orders.copy().items():
                for uuid in uuids:
                    order = self._mem_orders.get(uuid)
                    if order:
                        await cursor.execute(
                            "INSERT INTO open_orders (uuid, exchange, symbol, strategy_id, user_id) VALUES (?, ?, ?, ?, ?)",
                            (uuid, exchange.value, order.symbol, self.strategy_id, self.user_id)
                        )

            await self._db_async.commit()

    def _cleanup_expired_data(self):
        """Cleanup expired data"""
        current_time = self._clock.timestamp_ms()
        expire_before = current_time - self._expire_time * 1000

        # Clean up expired orders
        expired_orders = [
            uuid
            for uuid, order in self._mem_orders.copy().items()
            if order.timestamp < expire_before
        ]
        for uuid in expired_orders:
            del self._mem_orders[uuid]
            self._mem_closed_orders.pop(uuid, None)
            self._log.debug(f"removing order {uuid} from memory")
            for symbol, order_set in self._mem_symbol_orders.copy().items():
                self._log.debug(f"removing order {uuid} from symbol {symbol}")
                order_set.discard(uuid)
        
        expired_algo_orders = [
            uuid
            for uuid, algo_order in self._mem_algo_orders.copy().items()
            if algo_order.timestamp < expire_before
        ]
        for uuid in expired_algo_orders:
            del self._mem_algo_orders[uuid]
            self._log.debug(f"removing algo order {uuid} from memory")

    async def close(self):
        """Close the cache"""
        self._shutdown_event.set()
        await self._sync_to_redis()
        await self._r_async.aclose()
        await self._db_async.close()
        self._db.close()

    ################ # cache public data  ###################

    def _update_kline_cache(self, kline: Kline):
        self._kline_cache[kline.symbol] = kline

    def _update_bookl1_cache(self, bookl1: BookL1):
        self._bookl1_cache[bookl1.symbol] = bookl1

    def _update_trade_cache(self, trade: Trade):
        self._trade_cache[trade.symbol] = trade

    def kline(self, symbol: str) -> Optional[Kline]:
        """
        Retrieve a Kline object from the cache by symbol.

        :param symbol: The symbol of the Kline to retrieve.
        :return: The Kline object if found, otherwise None.
        """
        return self._kline_cache.get(symbol, None)

    def bookl1(self, symbol: str) -> Optional[BookL1]:
        """
        Retrieve a BookL1 object from the cache by symbol.

        :param symbol: The symbol of the BookL1 to retrieve.
        :return: The BookL1 object if found, otherwise None.
        """
        return self._bookl1_cache.get(symbol, None)

    def trade(self, symbol: str) -> Optional[Trade]:
        """
        Retrieve a Trade object from the cache by symbol.

        :param symbol: The symbol of the Trade to retrieve.
        :return: The Trade object if found, otherwise None.
        """
        return self._trade_cache.get(symbol, None)

    ################ # cache private data  ###################

    def _check_status_transition(self, order: Order):
        previous_order = self._mem_orders.get(order.uuid)
        if not previous_order:
            return True

        if order.status not in STATUS_TRANSITIONS[previous_order.status]:
            self._log.error(
                f"Order id: {order.uuid} Invalid status transition: {previous_order.status} -> {order.status}"
            )
            return False

        return True
    
    def _apply_position(self, position: Position):
        self._mem_positions[position.symbol] = position
    
    def _apply_balance(self, account_type: AccountType, balances: List[Balance]):
        self._mem_account_balance[account_type]._apply(balances)
    
    def get_balance(self, account_type: AccountType) -> AccountBalance:
        return self._mem_account_balance[account_type]
    
    
    def _get_position_from_redis(self, instrument_id: InstrumentId) -> Position | None:
        key = f"strategy:{self.strategy_id}:user_id:{self.user_id}:exchange:{instrument_id.exchange.value}:symbol_positions:{instrument_id.symbol}"
        if position_data := self._r.get(key):
            position = self._decode(position_data, Position)
            self._mem_positions[instrument_id.symbol] = position  # Cache in memory
            return position
        return None
    
    def _get_position_from_sqlite(self, instrument_id: InstrumentId) -> Position | None:
        try:
            cursor = self._db.cursor()
            cursor.execute(
                """
                SELECT data FROM positions 
                WHERE strategy_id = ? AND user_id = ? AND symbol = ?
                """,
                (
                    self.strategy_id,
                    self.user_id,
                    instrument_id.symbol,
                )
            )
            if row := cursor.fetchone():
                position = self._decode(row[0], Position)
                self._mem_positions[instrument_id.symbol] = position  # Cache in memory
                return position
            return None
        except sqlite3.Error as e:
            self._log.error(f"Error getting position from SQLite: {e}")
            return None
        
    @maybe
    def get_position(self, symbol: str) -> Optional[Position]:
        # First try memory
        instrument_id = InstrumentId.from_str(symbol)
        if position := self._mem_positions.get(symbol, None):
            return position
            
        if self._storage_backend == StorageBackend.REDIS:
            return self._get_position_from_redis(instrument_id)
        elif self._storage_backend == StorageBackend.SQLITE:
            return self._get_position_from_sqlite(instrument_id)
    
    
    def _get_all_positions_from_redis(self, exchange: ExchangeType) -> Dict[str, Position]:
        positions = {}
        key = f"strategy:{self.strategy_id}:user_id:{self.user_id}:exchange:{exchange.value}:symbol_positions:*"
        if keys := self._r.keys(key):
            for position_key in keys:
                symbol = position_key.decode().split(":")[-1]
                if symbol in positions:
                    continue
                if position_data := self._r.get(position_key):
                    position = self._decode(position_data, Position)
                    positions[symbol] = position
                    self._mem_positions[symbol] = position  # Cache in memory
        return positions

    def _get_all_positions_from_sqlite(self, exchange: ExchangeType) -> Dict[str, Position]:
        positions = {}
        cursor = self._db.cursor()
        cursor.execute(
            """
            SELECT symbol, data FROM positions WHERE strategy_id = ? AND user_id = ? AND exchange = ?
            """,
            (self.strategy_id, self.user_id, exchange.value)
        )
        for row in cursor.fetchall():
            symbol, position_data = row
            if symbol in self._mem_positions:
                continue
            position = self._decode(position_data, Position)
            positions[symbol] = position
            self._mem_positions[symbol] = position  # Cache in memory
        return positions
        
    def get_all_positions(self, exchange: ExchangeType) -> Dict[str, Position]:
        positions = {symbol: position for symbol, position in self._mem_positions.copy().items() if position.exchange == exchange}
        
        if self._storage_backend == StorageBackend.REDIS:
            positions.update(self._get_all_positions_from_redis(exchange))
        elif self._storage_backend == StorageBackend.SQLITE:
            positions.update(self._get_all_positions_from_sqlite(exchange))
        
        return positions

    def _order_initialized(self, order: Order | AlgoOrder):
        if isinstance(order, AlgoOrder):
            self._mem_algo_orders[order.uuid] = order
        else:
            if not self._check_status_transition(order):
                return
            self._mem_orders[order.uuid] = order
            self._mem_open_orders[order.exchange].add(order.uuid)
            self._mem_symbol_orders[order.symbol].add(order.uuid)
            self._mem_symbol_open_orders[order.symbol].add(order.uuid)

    def _order_status_update(self, order: Order | AlgoOrder):
        if isinstance(order, AlgoOrder):
            self._mem_algo_orders[order.uuid] = order
        else:
            if not self._check_status_transition(order):
                return
            self._mem_orders[order.uuid] = order
            if order.is_closed:
                self._mem_open_orders[order.exchange].discard(order.uuid)
                self._mem_symbol_open_orders[order.symbol].discard(order.uuid)

    
    def _get_order_from_redis(self, uuid: str) -> Optional[Order | AlgoOrder]:
        # find in memory first
        if uuid.startswith("ALGO-"):
            if order := self._mem_algo_orders.get(uuid):
                return order
            key = f"strategy:{self.strategy_id}:user_id:{self.user_id}:algo_orders"
            obj_type = AlgoOrder
            mem_dict = self._mem_algo_orders
        else:
            if order := self._mem_orders.get(uuid):
                return order
            key = f"strategy:{self.strategy_id}:user_id:{self.user_id}:orders"
            obj_type = Order
            mem_dict = self._mem_orders

        if raw_order := self._r.hget(key, uuid):
            order = self._decode(raw_order, obj_type)
            mem_dict[uuid] = order
            return order
        return None
    
    def _get_order_from_sqlite(self, uuid: str) -> Optional[Order | AlgoOrder]:
        try:
            # determine the table and object type to query
            if uuid.startswith("ALGO-"):
                table = "algo_orders"
                obj_type = AlgoOrder
                mem_dict = self._mem_algo_orders
            else:
                table = "orders"
                obj_type = Order
                mem_dict = self._mem_orders

            # find in memory
            if order := mem_dict.get(uuid):
                return order

            # query SQLite
            cursor = self._db.cursor()
            cursor.execute(
                f"""
                SELECT data FROM {table}
                WHERE uuid = ?
                """,
                (uuid, )
            )
            
            if row := cursor.fetchone():
                order = self._decode(row[0], obj_type)
                mem_dict[uuid] = order  # Cache in memory
                return order
                
            return None
            
        except sqlite3.Error as e:
            self._log.error(f"Error getting order from SQLite: {e}")
            return None
    
    @maybe
    def get_order(self, uuid: str) -> Optional[Order | AlgoOrder]:
        if self._storage_backend == StorageBackend.REDIS:
            return self._get_order_from_redis(uuid)
        elif self._storage_backend == StorageBackend.SQLITE:
            return self._get_order_from_sqlite(uuid)
    
    
    def _get_symbol_orders_from_redis(self, instrument_id: InstrumentId) -> Set[str]:
        key = f"strategy:{self.strategy_id}:user_id:{self.user_id}:exchange:{instrument_id.exchange.value}:symbol_orders:{instrument_id.symbol}"
        if redis_orders := self._r.smembers(key):
            return {uuid.decode() for uuid in redis_orders}
        return set()
    
    def _get_symbol_orders_from_sqlite(self, instrument_id: InstrumentId) -> Set[str]:
        cursor = self._db.cursor()
        cursor.execute(
            """
            SELECT uuid FROM orders WHERE strategy_id = ? AND user_id = ? AND symbol = ?
            """,
            (self.strategy_id, self.user_id, instrument_id.symbol)
        )
        return {row[0] for row in cursor.fetchall()}

    def get_symbol_orders(self, symbol: str, in_mem: bool = True) -> Set[str]:
        """Get all orders for a symbol from memory and Redis"""
        memory_orders = self._mem_symbol_orders.get(symbol, set())
        if not in_mem:
            instrument_id = InstrumentId.from_str(symbol)
            if self._storage_backend == StorageBackend.REDIS:
                orders = self._get_symbol_orders_from_redis(instrument_id)
            elif self._storage_backend == StorageBackend.SQLITE:
                orders = self._get_symbol_orders_from_sqlite(instrument_id)
            return memory_orders.union(orders)
        return memory_orders

    def get_open_orders(
        self, symbol: str | None = None, exchange: ExchangeType | None = None
    ) -> Set[str]:
        if symbol is not None:
            return self._mem_symbol_open_orders[symbol]
        elif exchange is not None:
            return self._mem_open_orders[exchange]
        else:
            raise ValueError("Either `symbol` or `exchange` must be specified")








import asyncio
import platform
from typing import Dict, Set
from nexustrader.constants import AccountType, ExchangeType
from nexustrader.config import Config
from nexustrader.strategy import Strategy
from nexustrader.core.cache import AsyncCache
from nexustrader.core.registry import OrderRegistry
from nexustrader.error import EngineBuildError, SubscriptionError
from nexustrader.base import (
    ExchangeManager,
    PublicConnector,
    PrivateConnector,
    ExecutionManagementSystem,
    OrderManagementSystem,
)
from nexustrader.exchange.bybit import (
    BybitExchangeManager,
    BybitPrivateConnector,
    BybitPublicConnector,
    BybitAccountType,
    BybitExecutionManagementSystem,
    BybitOrderManagementSystem,
)
from nexustrader.exchange.binance import (
    BinanceExchangeManager,
    BinanceAccountType,
    BinancePublicConnector,
    BinancePrivateConnector,
    BinanceExecutionManagementSystem,
    BinanceOrderManagementSystem,
)
from nexustrader.exchange.okx import (
    OkxExchangeManager,
    OkxAccountType,
    OkxPublicConnector,
    OkxPrivateConnector,
    OkxExecutionManagementSystem,
    OkxOrderManagementSystem,
)
from nexustrader.core.entity import TaskManager, ZeroMQSignalRecv
from nexustrader.core.nautilius_core import MessageBus, TraderId, LiveClock
from nexustrader.schema import InstrumentId
from nexustrader.constants import DataType


class Engine:
    @staticmethod
    def set_loop_policy():
        if platform.system() != "Windows":
            import uvloop

            asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

    def __init__(self, config: Config):
        self._config = config
        self._is_running = False
        self._is_built = False
        self._scheduler_started = False
        self.set_loop_policy()
        self._loop = asyncio.new_event_loop()
        self._task_manager = TaskManager(self._loop)

        self._exchanges: Dict[ExchangeType, ExchangeManager] = {}
        self._public_connectors: Dict[AccountType, PublicConnector] = {}
        self._private_connectors: Dict[AccountType, PrivateConnector] = {}

        trader_id = f"{self._config.strategy_id}-{self._config.user_id}"

        self._custom_signal_recv = None

        self._msgbus = MessageBus(
            trader_id=TraderId(trader_id),
            clock=LiveClock(),
        )
        self._cache: AsyncCache = AsyncCache(
            strategy_id=config.strategy_id,
            user_id=config.user_id,
            msgbus=self._msgbus,
            task_manager=self._task_manager,
            sync_interval=config.cache_sync_interval,
            expire_time=config.cache_expire_time,
        )

        self._registry = OrderRegistry()

        self._oms: Dict[ExchangeType, OrderManagementSystem] = {}
        self._ems: Dict[ExchangeType, ExecutionManagementSystem] = {}

        self._strategy: Strategy = config.strategy
        self._strategy._init_core(
            cache=self._cache,
            msgbus=self._msgbus,
            task_manager=self._task_manager,
            ems=self._ems,
            exchanges=self._exchanges,
            private_connectors=self._private_connectors,
        )

        self._subscriptions: Dict[DataType, Dict[str, str] | Set[str]] = (
            self._strategy._subscriptions
        )

    def _public_connector_check(self):
        okx_public_conn_count = 0

        for account_type in self._public_connectors.keys():
            if isinstance(account_type, BybitAccountType):
                if (
                    account_type == BybitAccountType.UNIFIED
                    or account_type == BybitAccountType.UNIFIED_TESTNET
                ):
                    raise EngineBuildError(
                        f"{account_type} is not supported for public connector."
                    )
                bybit_basic_config = self._config.basic_config.get(ExchangeType.BYBIT)
                if not bybit_basic_config:
                    raise EngineBuildError(
                        f"Basic config for {ExchangeType.BYBIT} is not set. Please add `{ExchangeType.BYBIT}` in `basic_config`."
                    )

                else:
                    if bybit_basic_config.testnet != account_type.is_testnet:
                        raise EngineBuildError(
                            f"The `testnet` setting of {ExchangeType.BYBIT} is not consistent with the public connector's account type `{account_type}`."
                        )
            elif isinstance(account_type, BinanceAccountType):
                if (
                    account_type.is_isolated_margin_or_margin
                    or account_type.is_portfolio_margin
                ):
                    raise EngineBuildError(
                        f"{account_type} is not supported for public connector."
                    )
                binance_basic_config = self._config.basic_config.get(
                    ExchangeType.BINANCE
                )
                if not binance_basic_config:
                    raise EngineBuildError(
                        f"Basic config for {ExchangeType.BINANCE} is not set. Please add `{ExchangeType.BINANCE}` in `basic_config`."
                    )
                else:
                    if binance_basic_config.testnet != account_type.is_testnet:
                        raise EngineBuildError(
                            f"The `testnet` setting of {ExchangeType.BINANCE} is not consistent with the public connector's account type `{account_type}`."
                        )

            elif isinstance(account_type, OkxAccountType):
                if okx_public_conn_count > 1:
                    raise EngineBuildError(
                        "Only one public connector is supported for OKX, please remove the extra public connector config."
                    )
                okx_basic_config = self._config.basic_config.get(ExchangeType.OKX)
                if not okx_basic_config:
                    raise EngineBuildError(
                        f"Basic config for {ExchangeType.OKX} is not set. Please add `{ExchangeType.OKX}` in `basic_config`."
                    )
                else:
                    if okx_basic_config.testnet != account_type.is_testnet:
                        raise EngineBuildError(
                            f"The `testnet` setting of {ExchangeType.OKX} is not consistent with the public connector's account type `{account_type}`."
                        )

                okx_public_conn_count += 1
            else:
                raise EngineBuildError(f"Unsupported account type: {account_type}")

    def _build_public_connectors(self):
        for exchange_id, public_conn_configs in self._config.public_conn_config.items():
            for config in public_conn_configs:
                if exchange_id == ExchangeType.BYBIT:
                    exchange: BybitExchangeManager = self._exchanges[exchange_id]
                    account_type: BybitAccountType = config.account_type
                    public_connector = BybitPublicConnector(
                        account_type=account_type,
                        exchange=exchange,
                        msgbus=self._msgbus,
                        task_manager=self._task_manager,
                    )
                    self._public_connectors[account_type] = public_connector

                elif exchange_id == ExchangeType.BINANCE:
                    exchange: BinanceExchangeManager = self._exchanges[exchange_id]
                    account_type: BinanceAccountType = config.account_type
                    public_connector = BinancePublicConnector(
                        account_type=account_type,
                        exchange=exchange,
                        msgbus=self._msgbus,
                        task_manager=self._task_manager,
                    )

                    self._public_connectors[account_type] = public_connector

                elif exchange_id == ExchangeType.OKX:
                    exchange: OkxExchangeManager = self._exchanges[exchange_id]
                    account_type: OkxAccountType = config.account_type
                    public_connector = OkxPublicConnector(
                        account_type=account_type,
                        exchange=exchange,
                        msgbus=self._msgbus,
                        task_manager=self._task_manager,
                    )
                    self._public_connectors[account_type] = public_connector
        self._public_connector_check()

    def _build_private_connectors(self):
        for (
            exchange_id,
            private_conn_configs,
        ) in self._config.private_conn_config.items():
            if not private_conn_configs:
                raise EngineBuildError(
                    f"Private connector config for {exchange_id} is not set. Please add `{exchange_id}` in `private_conn_config`."
                )

            match exchange_id:
                case ExchangeType.BYBIT:
                    config = private_conn_configs[0]
                    exchange: BybitExchangeManager = self._exchanges[exchange_id]
                    account_type = (
                        BybitAccountType.UNIFIED_TESTNET
                        if exchange.is_testnet
                        else BybitAccountType.UNIFIED
                    )

                    private_connector = BybitPrivateConnector(
                        exchange=exchange,
                        account_type=account_type,
                        cache=self._cache,
                        msgbus=self._msgbus,
                        rate_limit=config.rate_limit,
                        task_manager=self._task_manager,
                    )
                    self._private_connectors[account_type] = private_connector

                case ExchangeType.OKX:
                    assert (
                        len(private_conn_configs) == 1
                    ), "Only one private connector is supported for OKX, please remove the extra private connector config."

                    config = private_conn_configs[0]
                    exchange: OkxExchangeManager = self._exchanges[exchange_id]
                    account_type = (
                        OkxAccountType.DEMO if exchange.is_testnet else OkxAccountType.LIVE
                    )

                    private_connector = OkxPrivateConnector(
                        exchange=exchange,
                        account_type=account_type,
                        cache=self._cache,
                        msgbus=self._msgbus,
                        rate_limit=config.rate_limit,
                        task_manager=self._task_manager,
                    )
                    self._private_connectors[account_type] = private_connector

                case ExchangeType.BINANCE:
                    for config in private_conn_configs:
                        exchange: BinanceExchangeManager = self._exchanges[exchange_id]
                        account_type: BinanceAccountType = config.account_type

                        private_connector = BinancePrivateConnector(
                            exchange=exchange,
                            account_type=account_type,
                            cache=self._cache,
                            msgbus=self._msgbus,
                            rate_limit=config.rate_limit,
                            task_manager=self._task_manager,
                        )
                        self._private_connectors[account_type] = private_connector

    def _build_exchanges(self):
        for exchange_id, basic_config in self._config.basic_config.items():
            config = {
                "apiKey": basic_config.api_key,
                "secret": basic_config.secret,
                "sandbox": basic_config.testnet,
            }
            if basic_config.passphrase:
                config["password"] = basic_config.passphrase

            if exchange_id == ExchangeType.BYBIT:
                self._exchanges[exchange_id] = BybitExchangeManager(config)
            elif exchange_id == ExchangeType.BINANCE:
                self._exchanges[exchange_id] = BinanceExchangeManager(config)
            elif exchange_id == ExchangeType.OKX:
                self._exchanges[exchange_id] = OkxExchangeManager(config)

    def _build_custom_signal_recv(self):
        zmq_config = self._config.zero_mq_signal_config
        if zmq_config:
            if not hasattr(self._strategy, "on_custom_signal"):
                raise EngineBuildError(
                    "Please add `on_custom_signal` method to the strategy."
                )

            self._custom_signal_recv = ZeroMQSignalRecv(
                zmq_config, self._strategy.on_custom_signal, self._task_manager
            )

    def _build_ems(self):
        for exchange_id in self._exchanges.keys():
            match exchange_id:
                case ExchangeType.BYBIT:
                    exchange: BybitExchangeManager = self._exchanges[exchange_id]
                    self._ems[exchange_id] = BybitExecutionManagementSystem(
                        market=exchange.market,
                        cache=self._cache,
                        msgbus=self._msgbus,
                        task_manager=self._task_manager,
                        registry=self._registry,
                    )
                    self._ems[exchange_id]._build(self._private_connectors)
                case ExchangeType.BINANCE:
                    exchange: BinanceExchangeManager = self._exchanges[exchange_id]
                    self._ems[exchange_id] = BinanceExecutionManagementSystem(
                        market=exchange.market,
                        cache=self._cache,
                        msgbus=self._msgbus,
                        task_manager=self._task_manager,
                        registry=self._registry,
                    )
                    self._ems[exchange_id]._build(self._private_connectors)
                case ExchangeType.OKX:
                    exchange: OkxExchangeManager = self._exchanges[exchange_id]
                    self._ems[exchange_id] = OkxExecutionManagementSystem(
                        market=exchange.market,
                        cache=self._cache,
                        msgbus=self._msgbus,
                        task_manager=self._task_manager,
                        registry=self._registry,
                    )
                    self._ems[exchange_id]._build(self._private_connectors)

    def _build_oms(self):
        for exchange_id in self._exchanges.keys():
            match exchange_id:
                case ExchangeType.BYBIT:
                    self._oms[exchange_id] = BybitOrderManagementSystem(
                        cache=self._cache,
                        msgbus=self._msgbus,
                        task_manager=self._task_manager,
                        registry=self._registry,
                    )
                case ExchangeType.BINANCE:
                    self._oms[exchange_id] = BinanceOrderManagementSystem(
                        cache=self._cache,
                        msgbus=self._msgbus,
                        task_manager=self._task_manager,
                        registry=self._registry,
                    )
                case ExchangeType.OKX:
                    self._oms[exchange_id] = OkxOrderManagementSystem(
                        cache=self._cache,
                        msgbus=self._msgbus,
                        task_manager=self._task_manager,
                        registry=self._registry,
                    )

    def _build(self):
        self._build_exchanges()
        self._build_public_connectors()
        self._build_private_connectors()
        self._build_ems()
        self._build_oms()
        self._build_custom_signal_recv()
        self._is_built = True

    def _instrument_id_to_account_type(
        self, instrument_id: InstrumentId
    ) -> AccountType:
        match instrument_id.exchange:
            case ExchangeType.BYBIT:
                bybit_basic_config = self._config.basic_config.get(ExchangeType.BYBIT)
                if not bybit_basic_config:
                    raise EngineBuildError(
                        f"Basic config for {ExchangeType.BYBIT} is not set. Please add `{ExchangeType.BYBIT}` in `basic_config`."
                    )

                is_testnet = bybit_basic_config.testnet

                if instrument_id.is_spot:
                    return (
                        BybitAccountType.SPOT_TESTNET
                        if is_testnet
                        else BybitAccountType.SPOT
                    )
                elif instrument_id.is_linear:
                    return (
                        BybitAccountType.LINEAR_TESTNET
                        if is_testnet
                        else BybitAccountType.LINEAR
                    )
                elif instrument_id.is_inverse:
                    return (
                        BybitAccountType.INVERSE_TESTNET
                        if is_testnet
                        else BybitAccountType.INVERSE
                    )
                else:
                    raise ValueError(
                        f"Unsupported instrument type: {instrument_id.type}"
                    )
            case ExchangeType.BINANCE:
                binance_basic_config = self._config.basic_config.get(
                    ExchangeType.BINANCE
                )
                if not binance_basic_config:
                    raise EngineBuildError(
                        f"Basic config for {ExchangeType.BINANCE} is not set. Please add `{ExchangeType.BINANCE}` in `basic_config`."
                    )

                is_testnet = binance_basic_config.testnet

                if instrument_id.is_spot:
                    return (
                        BinanceAccountType.SPOT_TESTNET
                        if is_testnet
                        else BinanceAccountType.SPOT
                    )
                elif instrument_id.is_linear:
                    return (
                        BinanceAccountType.USD_M_FUTURE_TESTNET
                        if is_testnet
                        else BinanceAccountType.USD_M_FUTURE
                    )
                elif instrument_id.is_inverse:
                    return (
                        BinanceAccountType.COIN_M_FUTURE_TESTNET
                        if is_testnet
                        else BinanceAccountType.COIN_M_FUTURE
                    )
                else:
                    raise ValueError(
                        f"Unsupported instrument type: {instrument_id.type}"
                    )
            case ExchangeType.OKX:
                account_types = self._config.public_conn_config.get(ExchangeType.OKX)
                if not account_types:
                    raise EngineBuildError(
                        f"Public connector config for {ExchangeType.OKX} is not set. Please add `{ExchangeType.OKX}` in `public_conn_config`."
                    )

                return account_types[0].account_type

    async def _start_connectors(self):
        for connector in self._private_connectors.values():
            await connector.connect()

        for data_type, sub in self._subscriptions.items():
            match data_type:
                case DataType.BOOKL1:
                    for symbol in sub:
                        instrument_id = InstrumentId.from_str(symbol)
                        account_type = self._instrument_id_to_account_type(
                            instrument_id
                        )
                        connector = self._public_connectors.get(account_type, None)
                        if connector is None:
                            raise SubscriptionError(
                                f"Please add `{account_type}` public connector to the `config.public_conn_config`."
                            )
                        await connector.subscribe_bookl1(instrument_id.symbol)
                case DataType.TRADE:
                    for symbol in sub:
                        instrument_id = InstrumentId.from_str(symbol)
                        account_type = self._instrument_id_to_account_type(
                            instrument_id
                        )
                        connector = self._public_connectors.get(account_type, None)
                        if connector is None:
                            raise SubscriptionError(
                                f"Please add `{account_type}` public connector to the `config.public_conn_config`."
                            )
                        await connector.subscribe_trade(instrument_id.symbol)
                case DataType.KLINE:
                    for symbol, interval in sub.items():
                        instrument_id = InstrumentId.from_str(symbol)
                        account_type = self._instrument_id_to_account_type(
                            instrument_id
                        )
                        connector = self._public_connectors.get(account_type, None)
                        if connector is None:
                            raise SubscriptionError(
                                f"Please add `{account_type}` public connector to the `config.public_conn_config`."
                            )
                        await connector.subscribe_kline(instrument_id.symbol, interval)
                case DataType.MARK_PRICE:
                    pass  # TODO: implement
                case DataType.FUNDING_RATE:
                    pass  # TODO: implement
                case DataType.INDEX_PRICE:
                    pass  # TODO: implement

    async def _start_ems(self):
        for ems in self._ems.values():
            await ems.start()

    async def _start_oms(self):
        for oms in self._oms.values():
            await oms.start()

    def _start_scheduler(self):
        self._strategy._scheduler.start()
        self._scheduler_started = True

    async def _start(self):
        await self._cache.start()
        await self._start_oms()
        await self._start_ems()
        await self._start_connectors()
        if self._custom_signal_recv:
            await self._custom_signal_recv.start()
        self._start_scheduler()
        await self._task_manager.wait()

    async def _dispose(self):
        if self._scheduler_started:
            self._strategy._scheduler.shutdown()
        for connector in self._public_connectors.values():
            await connector.disconnect()
        for connector in self._private_connectors.values():
            await connector.disconnect()

        await self._task_manager.cancel()

    def start(self):
        self._build()
        self._is_running = True
        self._loop.run_until_complete(self._start())

    def dispose(self):
        self._loop.run_until_complete(self._dispose())
        self._loop.close()
